/*!
  @file rx_fltk_glcanvas.cpp
	
  @brief FLTKɂOpenGLEBhENX
 
  @author Makoto Fujisawa 
  @date   2011-09
*/

//#ifdef _DEBUG
//#pragma comment (lib, "libjpegd.lib")
//#pragma comment (lib, "libpngd.lib")
//#pragma comment (lib, "zlibd.lib")
//#else
//#pragma comment (lib, "libjpeg.lib")
//#pragma comment (lib, "libpng.lib")
//#pragma comment (lib, "zlib.lib")
//#endif

//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_fltk_glcanvas.h"
#include "rx_fltk_window.h"

#include "rx_time.h"
#include "rx_matrix.h"
#include "rx_quaternion.h"

// 摜
//#include "rx_jpeg.h"
#include "rx_png.h"
#include "rx_bitmap.h"

// ݒt@C
#include "rx_atom_ini.h"


//-----------------------------------------------------------------------------
// 萔Eϐ
//-----------------------------------------------------------------------------
const GLfloat RX_LIGHT0_POS[4] = {  2.0f, 4.0f, 1.0f, 0.0f };
const GLfloat RX_LIGHT1_POS[4] = { -1.0f, -10.0f, -1.0f, 0.0f };

const GLfloat RX_LIGHT_AMBI[4] = { 0.1f, 0.1f, 0.1f, 1.0f };
const GLfloat RX_LIGHT_DIFF[4] = { 0.9f, 0.9f, 0.9f, 1.0f };
const GLfloat RX_LIGHT_SPEC[4] = { 0.2f, 0.2f, 0.2f, 1.0f };

const GLfloat RX_GREEN[] = { 0.1f, 1.0f, 0.1f, 1.0f };
const GLfloat RX_RED[]   = { 1.0f, 0.1f, 0.1f, 1.0f };
const GLfloat RX_BLUE[]  = { 0.1f, 0.1f, 1.0f, 1.0f };
const GLfloat RX_WHITE[] = { 1.0f, 1.0f, 1.0f, 1.0f };

const GLfloat RX_FOV = 45.0f;

const Vec3 RX_VERT_COLOR[7] = { Vec3(1.0, 0.3, 0.3), 
								Vec3(0.3, 1.0, 0.3), 
								Vec3(0.3, 0.3, 1.0), 
								Vec3(1.0, 1.0, 0.3), 
								Vec3(0.3, 1.0, 1.0), 
								Vec3(1.0, 0.3, 1.0),
								Vec3(0.3, 0.3, 0.3) }; 


extern rxINI *g_pINI;	//!< ݒt@C

double g_fTBTran[3] = {0, 0, -1};	//!< _ړpgbN{[̕sړ(ݒt@Cւ̕ۑp)
double g_fTBQuat[4] = {1, 0, 0, 0};	//!< _ړpgbN{[̉](ݒt@Cւ̕ۑp)

extern vector<string> g_vDefaultFiles;	//!< ftHgǂݍ݃t@C

double g_fSpacing = 4;		//!< fvX}bṽTvOԊu
double g_fPrtRad = 0.018;	//!< p[eBN̔a
double g_fZmax = 0.11;		//!< ֊sƂȂfvX臒l
int g_iNfilter = 3;			//!< fvXl̃tB^TCY
int g_iNiters = 3;			//!< ֊s̔
int g_iNumPrts = 5000;		//!< p[eBN

#ifdef RX_USE_TIMER
rxTimerAvg g_Time;					//!< όvZԌv
#endif

//-----------------------------------------------------------------------------
// rxFlWindowNX̎
//-----------------------------------------------------------------------------

//! RXgN^
rxFlGLWindow::rxFlGLWindow(int x_, int y_, int w_, int h_, const char* l, void *parent)
	: Fl_Gl_Window(x_, y_, w_, h_, l), m_iWinW(w_), m_iWinH(h_)
{
	m_pParent = (rxFlWindow*)parent;
	m_bAnimation = false;
	resizable(this);
	end();

	// SSM
	m_pSSMCPU = 0;
	m_pSSMGPU = 0;
	m_pSSM = 0;

	m_iSelectedGrid = -1;

	m_iViewMesh = 1;
	m_iFiltering = 1;
	m_iUseGPU = 0;
	
	// `tO
	m_iDraw = 1;
	m_iPolyDraw = 4;			// |S`tO
	m_iDepthDraw = 1+2+4;	// fvX}bv`tO

	// ݒt@C
	if(g_pINI){
		g_pINI->Set("trackball", "tx",  &g_fTBTran[0],  0.0);
		g_pINI->Set("trackball", "ty",  &g_fTBTran[1],  0.0);
		g_pINI->Set("trackball", "tz",  &g_fTBTran[2], -5.0);
		g_pINI->Set("trackball", "q0",  &g_fTBQuat[0],  1.0);
		g_pINI->Set("trackball", "q1",  &g_fTBQuat[1],  0.0);
		g_pINI->Set("trackball", "q2",  &g_fTBQuat[2],  0.0);
		g_pINI->Set("trackball", "q3",  &g_fTBQuat[3],  0.0);

		g_pINI->Set("draw", "drawing", &m_iDraw, m_iDraw);
		g_pINI->Set("draw", "poly_drawing", &m_iPolyDraw, m_iPolyDraw);
		g_pINI->Set("draw", "depth_drawing", &m_iDepthDraw, m_iDepthDraw);

		g_pINI->Set("ssm", "pnum",  &g_iNumPrts, g_iNumPrts);
		g_pINI->Set("ssm", "h",  &g_fSpacing, g_fSpacing);
		g_pINI->Set("ssm", "r",  &g_fPrtRad, g_fPrtRad);
		g_pINI->Set("ssm", "zmax",  &g_fZmax, g_fZmax);
		g_pINI->Set("ssm", "n_filter",  &g_iNfilter, g_iNfilter);
		g_pINI->Set("ssm", "n_iters",  &g_iNiters, g_iNiters);
	}
}

//! fXgN^
rxFlGLWindow::~rxFlGLWindow()
{
	if(m_pSSMCPU) delete m_pSSMCPU;
	if(m_pSSM) delete m_pSSM;
}


/*! 
 * GL̏֐
 */
void rxFlGLWindow::InitGL(void)
{
	// MARK:InitGL
	m_fBGColor[0] = 1.0; m_fBGColor[1] = 1.0; m_fBGColor[2] = 1.0;
	m_iMouseButton = -1;

	CuSetDevice(0);

	cout << "OpenGL Ver. " << glGetString(GL_VERSION) << endl;

	glClearColor((GLfloat)m_fBGColor[0], (GLfloat)m_fBGColor[1], (GLfloat)m_fBGColor[2], 1.0f);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// ݒ
	glLightfv(GL_LIGHT0, GL_AMBIENT,  RX_LIGHT_AMBI);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  RX_LIGHT_DIFF);
	glLightfv(GL_LIGHT0, GL_SPECULAR, RX_LIGHT_SPEC);
	glLightfv(GL_LIGHT0, GL_POSITION, RX_LIGHT0_POS);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

	glShadeModel(GL_SMOOTH);

	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	// gbN{[p
	m_tbView.SetTranslation(g_fTBTran[0], g_fTBTran[1]);
	m_tbView.SetScaling(g_fTBTran[2]);
	m_tbView.SetQuaternion(g_fTBQuat);

	// p[eBNf[^
	InitParticles();
}

/*! 
 * TCYCxg֐
 * @param[in] w LoX(sNZ)
 * @param[in] h LoX(sNZ)
 */
void rxFlGLWindow::Resize(int w, int h)
{
	m_iWinW = w;
	m_iWinH = h;

	glViewport(0, 0, w, h);
	m_tbView.SetRegion(w, h);

	//// ϊs̐ݒ
	//glMatrixMode(GL_PROJECTION);
	//glLoadIdentity();
	//Projection();

	//// fr[ϊs̐ݒ
	//glMatrixMode(GL_MODELVIEW);
	//glLoadIdentity();

	string status_str = ToString(w)+" x "+ToString(h);
	m_pParent->SetStatusLabel(status_str);
}

/*!
 * eϊ
 */
void rxFlGLWindow::Projection(int w, int h)
{
	gluPerspective(RX_FOV, (float)w/(float)h, 0.2f, 1000.0f);
	//glOrtho(-1, 1, -1, 1, -1, 1);
}

/*!
 * _
 */
void rxFlGLWindow::InitView(void)
{
	m_tbView.Init();
	m_tbView.SetScaling(-1.5);
}

/*!
 * xyz`(x:,y:,z:)
 * @param[in] len ̒
 */
inline int DrawAxis(double len, double line_width = 5.0)
{
	glLineWidth((GLfloat)line_width);

	// x
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glEnd();

	// y
	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, len, 0.0);
	glEnd();

	// z
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, len);
	glEnd();

	return 1;
}

/*!
 * ̕`
 * @param[in] cen ̒S
 * @param[in] rad ̔a
 * @param[in] col `F(gUːF)
 */
inline void DrawSphere(const Vec3 &cen, const double &rad)
{
	glPushMatrix();
	glTranslatef(cen[0], cen[1], cen[2]);
	glRotated(90, 1.0, 0.0, 0.0);
	glutSolidSphere(rad, 20, 10);
	glPopMatrix();
}


/*!
 * "\n"܂܂string𕡐stringɕ
 * @param[in] 
 * @return 
 */
void DivideString(const string &org, vector<string> &div)
{
	size_t pos0 = 0, pos1 = 0;
	while(pos1 != string::npos){
		pos1 = org.find("\n", pos0);

		div.push_back(org.substr(pos0, pos1-pos0));

		pos0 = pos1+1;
	}
}

/*!
 * ʕ`敶̐
 * @return ʕ`p
 */
vector<string> rxFlGLWindow::SetDrawString(void)
{
	// MARK:SetDrawString
	vector<string> str;

	str.push_back("SSM : ");
	if(m_pSSM){
		str.back() << m_pSSM->GetNx() << " x " << m_pSSM->GetNy();
	}
	str.back() << ", h = " << g_fSpacing;
	str.back() << ", Nfilter = " << g_iNfilter;
	str.back() << ", Niters = " << g_iNiters;

	str.push_back("smooting : ");
	str.back() << "depth " << ((m_iFiltering & 0x01) ? "on" : "off");
	str.back() << ", silhouette " << ((m_iFiltering & 0x02) ? "on" : "off");
	str.back() << ", GPU : " << ((m_iUseGPU & 0x01) ? "on" : "off");

#ifdef RX_USE_TIMER
	str.push_back("time : ");
	string tstr;
	RXTIMER_STRING(tstr);
	DivideString(tstr, str);
#endif

	return str;
}

/*!
 * `
 * @param[in] cir_str zobt@
 * @param[in] static_str ÓIȕobt@
 * @param[in] w,h EBhETCY
 */
static void DrawStrings(vector<string> &static_str, int w, int h)
{
	// MRK:PrintD
	glDisable(GL_LIGHTING);
	//glColor3f(0.0, 0.0, 0.0);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, w, 0, h);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	float x0 = 5.0f;
	float y0 = h-20.0f;

	// ʏ㕔ɃX^eBbNȃeLXg
	for(int j = 0; j < (int)static_str.size(); ++j){
		glRasterPos2f(x0, y0);

		int size = (int)static_str[j].size();
		for(int i = 0; i < size; ++i){
			char ic = static_str[j][i];
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, ic);
		}

		y0 -= 20;
	}

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

/*!
 * ĕ`Cxg֐
 */
void rxFlGLWindow::Display(void)
{
	// MARK:Display
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	// ʕ`
	if(m_iDraw & RX_DRAW_PARAMS){
		glColor3d(0.5, 0.5, 1.0);
		DrawStrings(SetDrawString(), m_iWinW, m_iWinH);
	}

	// p[eBN`
	glViewport(0, 0, m_iWinW/2, m_iWinH);
	RenderPolygon(m_iWinW/2, m_iWinH);

	// fvX`
	glViewport(m_iWinW/2, 0, m_iWinW/2, m_iWinH);
	RenderDepthMap(m_iWinW/2, m_iWinH);


	glViewport(0, 0, m_iWinW, m_iWinH);

	//swap_buffers();
}


/*!
 * }EXCxg֐
 * @param[in] button }EX{^(GLUT_LEFT_BUTTON,GLUT_MIDDLE_BUTTON,GLUT_RIGHT_BUTTON)
 * @param[in] state }EX{^̏(GLUT_UP, GLUT_DOWN)
 * @param[in] x,y }EXW(XN[Wn)
 */
void rxFlGLWindow::Mouse(int button, int state, int x, int y)
{
	if(x < 0 || y < 0 || x > m_iWinW || y > m_iWinH) return;

	make_current();
	m_iKeyMod = (Fl::event_state(FL_SHIFT) ? 2 : (Fl::event_state(FL_CTRL) ? 3 : 1));
	
	if(button == FL_LEFT_MOUSE){
		if(state){	// {^_E
			if(x < m_iWinW/2){
				m_tbView.Start(x, y, m_iKeyMod);
			}
			else{
				x -= m_iWinW/2;
				y = m_iWinH-y;
				SelectGrid(x, y);
			}
		}
		else{		// {^Abv
			m_tbView.Stop(x, y);

			// ݒt@Cւ̕ۑpɃgbN{[pm
			m_tbView.GetTranslation(g_fTBTran);
			m_tbView.GetScaling(g_fTBTran[2]);
			m_tbView.GetQuaternion(g_fTBQuat);
		}
	}
	else if(button == FL_MIDDLE_MOUSE){
	}
	else if(button == FL_RIGHT_MOUSE){
	}
	redraw();
}

/*!
 * [VCxg֐(}EX{^܂܃hbO)
 * @param[in] x,y }EXW(XN[Wn)
 */
void rxFlGLWindow::Motion(int x, int y)
{
	if(x < 0 || y < 0 || x > m_iWinW || y > m_iWinH) return;

	make_current();
	if(x < m_iWinW/2){
		m_tbView.Motion(x, y);
	}
	else{
		x -= m_iWinW/2;
		y = m_iWinH-y;

		SelectGrid(x, y);
	}
	redraw();
}

/*!
 * [VCxg֐(}EX{^Ȃړ)
 * @param[in] x,y }EXW(XN[Wn)
 */
void rxFlGLWindow::PassiveMotion(int x, int y)
{
	if(x < 0 || y < 0 || x > m_iWinW || y > m_iWinH) return;

	//make_current();
	//redraw();
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void rxFlGLWindow::Keyboard(int key, int x, int y)
{
	make_current();

	// CL[
	m_iKeyMod = (Fl::event_state(FL_SHIFT) ? 2 : (Fl::event_state(FL_CTRL) ? 3 : 1));
	if(m_iKeyMod != 1) return;

	switch(key){
	case 'g':	// OpenGLLpXEBhEŃtXN[ɂ
		SwitchFullScreen();
		break;

	default:
		break;
	}

	redraw();
}


/*!
 * AChR[obN֐
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnIdle_s(void* x)
{
	((rxFlGLWindow*)x)->Idle();
}
void rxFlGLWindow::Idle(void)
{
	make_current();
	static rxQuaternion rot(Vec3(0.0, 1.0, 0.0), 0.2);
	double rot_ang = rot.GetAngle();
	Vec3 rot_axis = rot.GetAxis();
	m_tbView.AddRotation(rot_ang, rot_axis[0], rot_axis[1], rot_axis[2]);
	redraw();
}

/*!
 * ^C}[R[obN֐
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnTimer_s(void* x)
{
	((rxFlGLWindow*)x)->Timer();
}
void rxFlGLWindow::Timer(void)
{
	Idle();
	if(m_bAnimation){
		Fl::repeat_timeout(0.033, OnTimer_s, this);
	}
}


/*!
 * ACh֐ON/OFF
 * @param[in] on trueON, falseOFF
 */
void rxFlGLWindow::SwitchIdle(int on)
{
	m_bAnimation = (on == -1) ? !m_bAnimation : (on ? true : false);
	if(m_bAnimation){
		//Fl::add_timeout(0.033, OnTimer_s, this);
		Fl::add_idle(OnIdle_s, this);
	}
	else{
		Fl::remove_idle(OnIdle_s, this);
	}
}


/*!
 * tXN[/EBhE\̐؂ւ
 */
void rxFlGLWindow::SwitchFullScreen(void)
{
	static int is_fullscreen = 0;
	static int pos0[2] = { 0, 0 };
	static int win0[2] = { 500, 500 };
	if(is_fullscreen){
		fullscreen_off(pos0[0], pos0[1], win0[0], win0[1]);
		is_fullscreen = 0;
	}
	else{
		pos0[0] = x();
		pos0[1] = y();
		win0[0] = w();
		win0[1] = h();
		fullscreen();
		is_fullscreen = 1;
	}

}

/*!
 * pXt@C菜pX𒊏o
 * @param[in] path pX
 * @return tH_pX
 */
inline string GetFolderPath(const string &path)
{
    size_t pos1;
 
    pos1 = path.rfind('\\');
    if(pos1 != string::npos){
        return path.substr(0, pos1+1);
        
    }
 
    pos1 = path.rfind('/');
    if(pos1 != string::npos){
        return path.substr(0, pos1+1);
    }
 
    return "";
}


/*!
 * ݂̉ʕ`摜t@CƂĕۑ
 * @param[in] fn t@CpX
 */
void rxFlGLWindow::SaveDisplay(const string &fn)
{
	static int count = 0;

	int w_ = w();
	int h_ = h();
	int c_ = 4;

	make_current();
	unsigned char* data = new unsigned char[w_*h_*c_];

	glReadPixels(0, 0, w_, h_, GL_RGBA, GL_UNSIGNED_BYTE, data);

	// ㉺]
	int stride = w_*c_;
	for(int j = 0; j < h_/2; ++j){
		for(int i = 0; i < stride; ++i){
			unsigned char tmp = data[j*stride+i];
			data[j*stride+i] = data[(h_-j-1)*stride+i];
			data[(h_-j-1)*stride+i] = tmp;
		}
	}

	string ext = GetExtension(fn);
	if(ext == "bmp"){
		WriteBitmapFile(fn, data, w_, h_, c_, RX_BMP_WINDOWS_V3);
		cout << "saved the screen image to " << fn << endl;
		count++;
	}
	else{
		if(ext == "png"){
			WritePngFile(fn, data, w_, h_, c_);
		}
		else{
			WritePngFile(fn+".png", data, w_, h_, c_);
		}
		cout << "saved the screen image to " << fn << endl;
		count++;
	}

	delete [] data;
}


/*!
 * Cxgnh
 * @param[in] ev CxgID
 */
int rxFlGLWindow::handle(int e)
{
	switch(e){
	case FL_DND_ENTER:
	case FL_DND_RELEASE:
	case FL_DND_LEAVE:
	case FL_DND_DRAG:
	case FL_PASTE:
		// DnDBoxɃy[Xgn߂ɁC
		// ̃CxgƂFl_Gl_Window::handleɏnȂD
		return 1;

	default:
		break;
	}

	return Fl_Gl_Window::handle(e);
}


/*!
 * Fl_Menu_Bar̃R[obN֐ - Draw
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnMenuDraw_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Draw/";
	label = label.substr(menu_name.size(), string::npos);

	((rxFlGLWindow*)x)->OnMenuDraw(1.0, label);
}
void rxFlGLWindow::OnMenuDraw(double val, string label)
{
	int idx = 0;
	while(label.find(RX_DRAW_STR[idx*2]) != 0) idx++;

	int flag = (0x01 << idx);
	m_iDraw ^= flag;
		
	m_pParent->UpdateMenuState();
	redraw();
}

/*!
 * Fl_Menu_Bar̃R[obN֐ - Polygon
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnMenuDrawPolygon_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Polygon/";
	label = label.substr(menu_name.size(), string::npos);

	((rxFlGLWindow*)x)->OnMenuDrawPolygon(1.0, label);
}
void rxFlGLWindow::OnMenuDrawPolygon(double val, string label)
{
	int idx = 0;
	while(label.find(RX_DRAW_POLY_STR[idx*2]) != 0) idx++;

	int flag = (0x01 << idx);
	m_iPolyDraw ^= flag;
		
	m_pParent->UpdateMenuState();
	redraw();
}

/*!
 * Fl_Menu_Bar̃R[obN֐ - Depth Map
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnMenuDrawDepthMap_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Depth Map/";
	label = label.substr(menu_name.size(), string::npos);

	((rxFlGLWindow*)x)->OnMenuDrawDepthMap(1.0, label);
}
void rxFlGLWindow::OnMenuDrawDepthMap(double val, string label)
{
	int idx = 0;
	while(label.find(RX_DRAW_DMAP_STR[idx*2]) != 0) idx++;

	int flag = (0x01 << idx);
	m_iDepthDraw ^= flag;
		
	m_pParent->UpdateMenuState();
	redraw();
}


/*!
 * Fl_Menu_Bar̃R[obN֐ - Draw
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnMenuSSM_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "SSM/";
	label = label.substr(menu_name.size(), string::npos);

	((rxFlGLWindow*)x)->OnMenuSSM(1.0, label);
}
void rxFlGLWindow::OnMenuSSM(double val, string label)
{
	if(label.find("Generate") != string::npos){	
		// bV
		CreateSSM();
	}
	else if(label.find("Interactive Generation") != string::npos){
		// A^CbV
		m_iViewMesh ^= 1;
	}
	else if(label.find("GPU") != string::npos){
		// GPUgp̐ؑ
		m_iUseGPU ^= 0x01;
		RXTIMER_CLEAR;
		if(m_iUseGPU){
			m_pSSM = m_pSSMGPU;
		}
		else{
			m_pSSM = m_pSSMCPU;
		}
		CreateSSM();
	}
	else if(label.find("Depth Filtering") != string::npos){
		// fvX}bvtB^O
		m_iFiltering ^= 0x01;
		CreateSSM();
	}
	else if(label.find("Silhouette Filtering") != string::npos){
		// ֊stB^O
		m_iFiltering ^= 0x02;
		CreateSSM();
	}
	else if(label.find("Selected Grid Info") != string::npos){
		// Obh̕\
		OutputGridInfo(m_iSelectedGrid);
	}
	else if(label.find("Reset Particles") != string::npos){
		// p[eBN
		InitParticles();
	}
	else if(label.find("Reset View") != string::npos){
		// _
		InitView();
	}
	else if(label.find("Reset Timer") != string::npos){
		// ^C}[Zbg
		RXTIMER_CLEAR;
	}

	m_pParent->UpdateMenuState();
	redraw();
}


/*!
 * Fl_Spinner̃R[obN֐ - Simulation
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlGLWindow::OnSpinSSM_s(Fl_Widget *widget, void* x)
{
	Fl_Spinner *spin = (Fl_Spinner*)widget;
	string label = spin->label();
	double val = spin->value();

	((rxFlGLWindow*)x)->OnSpinSSM(val, label);
}
void rxFlGLWindow::OnSpinSSM(double val, string label)
{
	if(label.find("Spacing") != string::npos){
		g_fSpacing = val;
		if(m_pSSM) m_pSSM->SetSpacing(g_fSpacing);
	}
	else if(label.find("Rad") != string::npos){
		g_fPrtRad = val;
		if(m_pSSM) m_pSSM->SetRadius(g_fPrtRad);
	}
	else if(label.find("Zmax") != string::npos){
		g_fZmax = val;
		if(m_pSSM) m_pSSM->SetZMax(g_fZmax);
	}
	else if(label.find("Nfilter") != string::npos){
		g_iNfilter = (int)(val+0.01);
		if(m_pSSM) m_pSSM->SetFilterRadius(g_iNfilter);
		cout << "Nfilter : " << g_iNfilter << endl;
	}
	else if(label.find("Niters") != string::npos){
		g_iNiters = (int)(val+0.01);
		if(m_pSSM) m_pSSM->SetSmoothIter(g_iNiters);
		cout << "Niters : " << g_iNiters << endl;
	}
	else if(label.find("Particles") != string::npos){
		g_iNumPrts = (int)(val+0.01);
		InitParticles();
	}

	CreateSSM(0);
	redraw();
}


//-----------------------------------------------------------------------------
// MARK:SSM
//-----------------------------------------------------------------------------
/*!
 * bVeXgpp[eBNf[^̍쐬
 */
void rxFlGLWindow::InitParticles(void)
{
	// eXgpf[^̍쐬
	srand(1234);

	m_vPrts.resize(g_iNumPrts);
	for(int i = 0; i < g_iNumPrts; ++i){
		m_vPrts[i] = RXFunc::Rand(Vec3(-0.2), Vec3(0.2));
		//m_vPrts[i] = RXFunc::Rand(Vec3(0.1, -0.2, -0.2), Vec3(0.2));
		//if(RXFunc::Rand(0.0, 1.0) <= 0.5){
		//	m_vPrts[i][0] *= -1;
		//}
	}
}

/*!
 * fvXl摜ŕۑ
 * @param[in] fn ۑt@C
 * @param[in] nx,ny fvX}bv̉𑜓x
 * @param[in] dm fvX}bv
 */
void rxFlGLWindow::SaveDepthMap(const string fn)
{
	int nx, ny;
	RXREAL *dm;

	m_pSSM->GetDepthMapSize(nx, ny);
	dm = m_pSSM->GetDepthMap();

	vector<unsigned char> img;
	img.resize(nx*ny*3);
	
	for(int j = 0; j < ny; ++j){
		for(int i = 0; i < nx; ++i){
			unsigned int ud = RX_DEPTH2COLOR(dm[j*nx+i]);

			int j1 = ny-j-1;
			img[3*(j1*nx+i)+0] = ud;
			img[3*(j1*nx+i)+1] = ud;
			img[3*(j1*nx+i)+2] = ud;
		}
	}

	WritePngFile(fn, &img[0], nx, ny, 3);
}


/*!
 * Screen Space Mesh̍쐬eXg
 */
void rxFlGLWindow::CreateSSM(int debug_output)
{
	// MARK:CreateSSM

	if(!m_iViewMesh){
		// OpenGLeϊs
		cout << "projection matrix" << endl;
		for(int j = 0; j < 4; ++j){
			for(int i = 0; i < 4; ++i){
				cout << " " << m_fProjectionMatrix[i*4+j];
			}
			cout << endl;
		}
		cout << endl;

		// OpenGLfr[ϊs
		cout << "modelview matrix" << endl;
		for(int j = 0; j < 4; ++j){
			for(int i = 0; i < 4; ++i){
				cout << " " << m_fModelviewMatrix[i*4+j];
			}
			cout << endl;
		}
		cout << endl;
	}

	
	if(m_vPrts.empty()){
		InitParticles();
	}

	if(!m_pSSM){
		m_pSSMCPU = new rxSSMeshCPU(g_fZmax, g_fSpacing, g_fPrtRad, g_iNfilter, g_iNiters);
		m_pSSMGPU = new rxSSMeshGPU(g_fZmax, g_fSpacing, g_fPrtRad, g_iNfilter, g_iNiters);
		if(m_iUseGPU){
			m_pSSM = m_pSSMGPU;
		}
		else{
			m_pSSM = m_pSSMCPU;
		}
	}
	else{
		m_pSSM->SetSpacing(g_fSpacing);
		m_pSSM->SetRadius(g_fPrtRad);
		m_pSSM->SetZMax(g_fZmax);
		m_pSSM->SetFilterRadius(g_iNfilter);
		m_pSSM->SetSmoothIter(g_iNiters);
	}

	if(!m_Poly.vertices.empty()){
		m_Poly.vertices.clear();
		m_Poly.normals.clear();
		m_Poly.faces.clear();
		m_Poly.materials.clear();
	}

	RXTIMER_RESET;

	// XN[Xy[XbV
	vector< vector<int> > tris;
	m_pSSM->CreateMesh(m_fProjectionMatrix, m_fModelviewMatrix, m_iWinW/2, m_iWinH, 
					   m_vPrts, g_iNumPrts, m_Poly.vertices, m_Poly.normals, tris, 
					   m_iFiltering, debug_output);

	RXTIMER_RESET;

	int n = (int)tris.size();
	m_Poly.faces.resize(n);
	for(int i = 0; i < n; ++i){
		m_Poly.faces[i].vert_idx.resize(3);

		for(int j = 0; j < 3; ++j){
			m_Poly.faces[i][j] = tris[i][j];
		}
	}
	if(m_Poly.normals.empty()){
		CalVertexNormals(m_Poly);
	}

	rxMaterialOBJ mat;
	mat.name      = "";
	mat.diffuse   = Vec4(0.2, 0.2, 1.0, 1.0);
	mat.specular  = Vec4(0.3, 0.3, 0.3, 1.0);
	mat.ambient   = Vec4(0.1, 0.1, 0.1, 1.0);
	mat.color     = Vec4(0.0);
	mat.emission  = Vec4(0.0);
	mat.shininess = 10.0;
	mat.illum = 2;
	mat.tex_file = "";
	mat.tex_name = 0;

	m_Poly.materials[mat.name] = mat;

	RXTIMER("polygon");

	int nx, ny;
	RXREAL *dm;

	m_pSSM->GetDepthMapSize(nx, ny);
	dm = m_pSSM->GetDepthMap();

	// fvXleNX`
	m_texDepth.SetSize(nx-1, ny-1, 4);
	for(int j = 0; j < ny-1; ++j){
		for(int i = 0; i < nx-1; ++i){
			//rxSSDepth depth = 0.25*(dm[i+j*nx]+dm[(i+1)+j*nx]+dm[i+(j+1)*nx]+dm[(i+1)+(j+1)*nx]);
			//unsigned int ud = RX_DEPTH2COLOR(depth);
			unsigned int ud = RX_DEPTH2COLOR(dm[i+j*nx]);

			m_texDepth.SetColor(i, j, ud, ud, ud, 255);
		}
	}
	BindTexture(m_texDepth);

	RXTIMER("texture");

}


/*!
 * }EXWl烁bVObh̑I
 * @param[in] x,y }EXWl
 */
void rxFlGLWindow::SelectGrid(int x, int y)
{
	int nx = m_pSSM->GetNx();
	int ny = m_pSSM->GetNy();
	int gx = (int)((double)x/(0.5*m_iWinW)*nx);
	int gy = (int)((double)y/(m_iWinH)*ny);

	int grid = gx+gy*nx;
	if(m_iSelectedGrid != grid){
		m_iSelectedGrid = grid;
		cout << "grid : " << gx << ", " << gy << endl;

		if(m_pSSM) m_pSSM->OutputGridInfo(grid);

		m_vSelectedMeshes.clear();
		rxSSGrid g = m_pSSM->GetMeshGrid(grid);
		m_vSelectedMeshes.resize(g.mesh_num);
		for(int i = 0; i < g.mesh_num; ++i){
			m_vSelectedMeshes[i] = g.mesh[i];
		}

	}
}


/*!
 * ObhɊւ̏o
 * @param[in] grid ObhCfbNX(i+j*nx)
 */
void rxFlGLWindow::OutputGridInfo(int grid)
{
	if(grid < 0){
		cout << "no grid is selected." << endl;
		return;
	}

	if(m_pSSM) m_pSSM->OutputGridVertexInfo(grid);

	// bV
	rxSSGrid g = m_pSSM->GetMeshGrid(grid);
	for(int i = 0; i < g.mesh_num; ++i){
		rxFace *face = &m_Poly.faces[g.mesh[i]];
		int n = (int)face->vert_idx.size();
	}
	cout << endl;
}


/*!
 * fvX`
 */
void rxFlGLWindow::RenderDepthMap(int w, int h)
{
	// MARK:RenderDepth
	if(!m_pSSM) return;

	Vec2 cnvs[2] = { Vec2(0.0, 0.0), Vec2(w, h) };

	// ϊs̐ݒ
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(cnvs[0][0], cnvs[1][0], cnvs[0][1], cnvs[1][1], -1, 1);

	// fr[ϊs̐ݒ
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

	int nx = m_pSSM->GetNx();
	int ny = m_pSSM->GetNy();

	// ֊sGbW
	if(m_iDepthDraw & RX_DRAW_DMAP_CONTOUR){
		glColor3d(1.0, 1.0, 0.0);
		glLineWidth(3.0);
		m_pSSM->DrawSilhouetteEdge();
	}

	// GbW
	if(m_iDepthDraw & RX_DRAW_DMAP_EDGE){
		glLineWidth(3.0);
		m_pSSM->DrawSSEdge();
	}

	// bV
	if((m_iDepthDraw & RX_DRAW_DMAP_FACE) && !m_Poly.faces.empty()){
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glColor3d(0.0, 1.0, 0.0);
		rxSSVertex *ssvrts = m_pSSM->GetSSVertex();
		glBegin(GL_TRIANGLES);
		for(int i = 0; i < (int)m_Poly.faces.size(); ++i){
			for(int j = 0; j < 3; ++j){
				int k = m_Poly.faces[i][j];
				Vec3 x = ssvrts[k].pos;
				glVertex3d(x[0], x[1], 0.01);
			}
		}
		glEnd();
	}

	// _ in screen space
	if((m_iDepthDraw & RX_DRAW_DMAP_VERTEX) && !m_Poly.vertices.empty()){
		glPointSize(5.0);
		m_pSSM->DrawSSVertex(Vec3(1.0, 0.5, 0.5), Vec3(0.5, 1.0, 0.5));
	}

	// IObh
	if(m_iSelectedGrid != -1){
		glColor3d(0.0, 0.0, 1.0);
		glLineWidth(1.0);
		int gx = m_iSelectedGrid%nx;
		int gy = m_iSelectedGrid/ny;
		double dw = (double)(w)/(double)(nx);
		double dh = (double)(h)/(double)(ny);

		glBegin(GL_LINE_LOOP);
		glVertex3d(dw*(gx), dh*(gy), 0.01);
		glVertex3d(dw*(gx+1), dh*(gy), 0.01);
		glVertex3d(dw*(gx+1), dh*(gy+1), 0.01);
		glVertex3d(dw*(gx), dh*(gy+1), 0.01);
		glEnd();

		m_pSSM->DrawMeshGrid(m_iSelectedGrid, RX_VERT_COLOR);

		// Obh̃bV
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glLineWidth(2.0);
		glColor3d(1.0, 0.3, 0.3);
		rxSSVertex *ssvrts = m_pSSM->GetSSVertex();
		glBegin(GL_TRIANGLES);
		for(int i = 0; i < (int)m_vSelectedMeshes.size(); ++i){
			int f = m_vSelectedMeshes[i];
			for(int j = 0; j < 3; ++j){
				int k = m_Poly.faces[f][j];
				Vec3 x = ssvrts[k].pos;
				glVertex3d(x[0], x[1], 0.02);
			}
		}
		glEnd();
	}

	glLineWidth(1.0);
	glPointSize(1.0);

	// fvXl
	if(m_iDepthDraw & RX_DRAW_DMAP_MAP){
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		if(m_texDepth.IsValid()){
			m_texDepth.Enable();
			m_texDepth.Bind();
		}

		glColor3d(0.0, 0.0, 0.0);
		glNormal3d(0.0, 0.0, 1.0);
		glBegin(GL_POLYGON);
		glTexCoord2d(0.0, 0.0); glVertex3d(cnvs[0][0], cnvs[0][1], 0.0);
		glTexCoord2d(0.0, 1.0); glVertex3d(cnvs[0][0], cnvs[1][1], 0.0);
		glTexCoord2d(1.0, 1.0); glVertex3d(cnvs[1][0], cnvs[1][1], 0.0);
		glTexCoord2d(1.0, 0.0); glVertex3d(cnvs[1][0], cnvs[0][1], 0.0);
		glEnd();

		if(m_texDepth.IsValid()){
			m_texDepth.Disable();
		}
	}
}

/*!
 * doubleglMaterial
 * @param[in] face ގw肷(GL_FRONT,GL_BACK,GL_FRONT_AND_BACK)
 * @param[in] pname w肷ގ̎(GL_AMBIENT,GL_DIFFUSE,GL_SPECULAR,GL_EMISSION,GL_SHININESS,GL_AMBIENT_AND_DIFFUSE,GL_COLOR_INDEXES)
 * @param[in] params ݒ肷ގ̒l
 * @return 
 */
void glMaterialdv(GLenum face, GLenum pname, const GLdouble *params)
{
	GLfloat col[4];
	col[0] = (GLfloat)params[0];
	col[1] = (GLfloat)params[1];
	col[2] = (GLfloat)params[2];
	col[3] = (GLfloat)params[3];
	glMaterialfv(face, pname, col);
}


/*!
 * |S̕`
 * @param[in] polys |Sf[^
 * @param[in] draw `tO(ʃrbg璸_,GbW,,@
 */
void rxFlGLWindow::DrawPolygons(rxPolygons &polys, int draw)
{
	// _ƃ|S
	int vn = (int)polys.vertices.size();
	int pn = (int)polys.faces.size();

	if(draw & RX_DRAW_POLY_FACE){
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_LIGHTING);
		glColor3d(0.0, 0.0, 1.0);
		bool use_mat = true;
		if(polys.materials.empty()){
			// ގݒ肳ĂȂĆCGbŴݕ`̏ꍇCGL_COLOR_MATERIALp
			use_mat = false;

			glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
			glEnable(GL_COLOR_MATERIAL);
		}
		else{
			glDisable(GL_COLOR_MATERIAL);
		}


		bool tc = false;
		rxMaterialOBJ *cur_mat = NULL, *pre_mat = NULL;

		// ׂẴ|S`
		for(int i = 0; i < pn; ++i){
			rxFace *face = &(polys.faces[i]);
			int n = (int)face->vert_idx.size();

			if(use_mat){
				// ގ̐ݒ
				cur_mat = &polys.materials[face->material_name];

				// ݂̍ގÔƈقȂꍇ̂݁COpenGL̍ގXV
				if(cur_mat != pre_mat){
					glColor4dv(cur_mat->color);
					glMaterialdv(GL_FRONT_AND_BACK, GL_DIFFUSE,  cur_mat->diffuse.data);
					glMaterialdv(GL_FRONT_AND_BACK, GL_SPECULAR, cur_mat->specular.data);
					glMaterialdv(GL_FRONT_AND_BACK, GL_AMBIENT,  cur_mat->ambient.data);
					//glMaterialdv(GL_FRONT_AND_BACK, GL_EMISSION, cur_mat->emission.data);
					//glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)cur_mat->shininess);

					// eNX`΃oCh
					if(cur_mat->tex_name){
						glEnable(GL_TEXTURE_2D);
						glBindTexture(GL_TEXTURE_2D, cur_mat->tex_name);
						glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					}
					else{
						glBindTexture(GL_TEXTURE_2D, 0);
						glDisable(GL_TEXTURE);
					}

					pre_mat = cur_mat;
				}
			}

			// |S`
			glBegin(GL_POLYGON);
			for(int j = 0; j < n; ++j){
				int idx = face->vert_idx[j];
				if(idx >= 0 && idx < vn){
					glNormal3dv(polys.normals[idx].data);
					if(!face->texcoords.empty()) glTexCoord2dv(face->texcoords[j].data);
					glVertex3dv((polys.vertices[idx]).data);
				}
			}
			glEnd();
		}

		glBindTexture(GL_TEXTURE_2D, 0);
		glDisable(GL_TEXTURE_2D);
	}


	// _`
	if(draw & RX_DRAW_POLY_VERTEX){
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);

		glDisable(GL_LIGHTING);
		glColor3d(1.0, 0.0, 0.0);
		glPointSize(5.0);
		glBegin(GL_POINTS);
		for(int i = 0; i < vn; ++i){
			glVertex3dv(polys.vertices[i].data);
		}
		glEnd();
	}

	// GbW`
	if(draw & RX_DRAW_POLY_EDGE){
		// GbW`ɂ"stitching"Ȃ߂̃ItZbg̐ݒ
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);

		glDisable(GL_LIGHTING);
		glColor3d(0.3, 1.0, 0.0);
		glLineWidth(1.0);
		for(int i = 0; i < pn; ++i){
			rxFace *face = &polys.faces[i];
			int n = (int)face->vert_idx.size();
			
			glBegin(GL_LINE_LOOP);
			for(int j = 0; j < n; ++j){
				int idx = face->vert_idx[j];
				if(idx >= 0 && idx < vn){
					glVertex3dv(polys.vertices[idx].data);
				}
			}
			glEnd();
		}

		// IObh̃bV
		glLineWidth(3.0);
		glColor3d(1.0, 0.0, 0.0);
		for(int i = 0; i < (int)m_vSelectedMeshes.size(); ++i){
			rxFace *face = &polys.faces[m_vSelectedMeshes[i]];
			int n = (int)face->vert_idx.size();

			glBegin(GL_LINE_LOOP);
			for(int j = 0; j < n; ++j){
				int idx = face->vert_idx[j];
				if(idx >= 0 && idx < vn){
					glVertex3dv(polys.vertices[idx].data);
				}
			}
			glEnd();
		}
	}
	//glDisable(GL_POLYGON_OFFSET_FILL);

	// @`
	if(draw & RX_DRAW_POLY_NORMAL){
		glDisable(GL_LIGHTING);
		glColor3d(0.0, 0.9, 0.0);
		glLineWidth(1.0);
		for(int i = 0; i < vn; ++i){
			glBegin(GL_LINES);
			glVertex3dv(polys.vertices[i].data);
			glVertex3dv((polys.vertices[i]+0.02*polys.normals[i]).data);
			glEnd();
		}
	}
}


/*!
 * SSMŐ|S̕`
 * @param[in] w,h `LoXTCY
 */
void rxFlGLWindow::RenderPolygon(int w, int h)
{
	// ϊs̐ݒ
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	Projection(w, h);

	// fr[ϊs̐ݒ
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glPushMatrix();
	m_tbView.Apply();	// }EXɂ]Esړ̓Kp

	glEnable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// OpenGLeϊs擾
	glGetDoublev(GL_PROJECTION_MATRIX, m_fProjectionMatrix);

	// OpenGLfr[ϊs擾
	glGetDoublev(GL_MODELVIEW_MATRIX, m_fModelviewMatrix);

	// bVȂΐ
	if(m_Poly.faces.empty() || m_iViewMesh){
		CreateSSM(0);
	}

	// p[eBN`
	if((m_iPolyDraw & RX_DRAW_POLY_PARTICLE) && !m_vPrts.empty()){
		glColor3f(0.0, 0.0, 1.0);
		for(int i = 0; i < (int)m_vPrts.size(); ++i){
			DrawSphere(m_vPrts[i], g_fPrtRad);
		}
	}

	// bV
	if(!m_Poly.faces.empty()){
		DrawPolygons(m_Poly, m_iPolyDraw);
	}

	// `
	if(m_iDraw & RX_DRAW_AXIS){
		glDisable(GL_LIGHTING);
		DrawAxis(1.0, 1.0);
	}

	glPopMatrix();
}