/*!
  @file rx_fltk_glcanvas.h
	
  @brief FLTKɂOpenGLEBhENX
 
  @author Makoto Fujisawa 
  @date   2011-09
*/

#ifndef _RX_FLTK_GLCANVAS_H_
#define _RX_FLTK_GLCANVAS_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <iostream>

// STL
#include <vector>
#include <string>

// FLTK
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Menu_Bar.H>
#include "rx_fltk_widgets.h"

// gbN{[
#include "rx_trackball.h"

// bV
#include "rx_mesh.h"

// SSM
#include "rx_ssm.h"
#include "rx_ssm.cuh"

// eNX`
#include "rx_texture.h"


using namespace std;

//-----------------------------------------------------------------------------
// MARK:`/萔
//-----------------------------------------------------------------------------
class rxFlWindow;

// `tO
enum
{
	RX_DRAW_PARAMS		= 0x0001,	//!< p[^`
	RX_DRAW_AXIS		= 0x0002,	//!< `
};
const string RX_DRAW_STR[] = {
	"Parameters",	"t", 
	"Axis",			"a", 
	"-1"
};

// |S`tO
enum
{
	RX_DRAW_POLY_VERTEX		= 0x0001,	//!< _`
	RX_DRAW_POLY_EDGE		= 0x0002,	//!< GbW`
	RX_DRAW_POLY_FACE		= 0x0004,	//!< ʕ`
	RX_DRAW_POLY_NORMAL		= 0x0008,	//!< @`
	RX_DRAW_POLY_PARTICLE	= 0x0010,	//!< p[eBN`
};
const string RX_DRAW_POLY_STR[] = {
	"Vertex",	"v", 
	"Edge",		"e", 
	"Face",		"f", 
	"Normal",	"n", 
	"Particle",	"p", 
	"-1"
};

// fvX}bv`tO
enum
{
	RX_DRAW_DMAP_VERTEX		= 0x0001,	//!< _`
	RX_DRAW_DMAP_CONTOUR	= 0x0002,	//!< ֊sObhGbW`
	RX_DRAW_DMAP_EDGE		= 0x0010,	//!< ObhGbW`
	RX_DRAW_DMAP_FACE		= 0x0004,	//!< bV`
	RX_DRAW_DMAP_MAP		= 0x0008,	//!< }bv`
};
const string RX_DRAW_DMAP_STR[] = {
	"Vertex",	"", 
	"Contour",	"", 
	"Edge",		"", 
	"Mesh",		"", 
	"Map",		"", 
	"-1"
};



//-----------------------------------------------------------------------------
//! rxFlGLWindowNX - fltkɂOpenGLEBhE
//-----------------------------------------------------------------------------
class rxFlGLWindow : public Fl_Gl_Window
{
protected:
	int m_iWinW;			//!< `EBhE̕
	int m_iWinH;			//!< `EBhE̍
	int m_iMouseButton;		//!< }EX{^̏
	int m_iKeyMod;			//!< CL[̏
	rxTrackball m_tbView;	//!< gbN{[

	double m_fBGColor[3];	//!< wiF
	bool m_bAnimation;		//!< Aj[VON/OFF

	rxFlWindow *m_pParent;	//!< eNX

	//
	// Screen Space Meshpϐ
	//
	rxSSMeshCPU *m_pSSMCPU;			//!< Screen Space Mesh by CPU
	rxSSMeshGPU *m_pSSMGPU;			//!< Screen Space Mesh by GPU
	rxSSMesh *m_pSSM;				//!< Screen Space Mesh

	vector<Vec3> m_vPrts;			//!< eXgpp[eBN
	rxPolygons m_Poly;				//!< bV

	double m_fProjectionMatrix[16];	//!< es
	double m_fModelviewMatrix[16];	//!< fr[s

	string m_strTime;				//!< Ԍv

	int m_iSelectedGrid;			//!< `pIObhCfbNX
	vector<int> m_vSelectedMeshes;	//!< IObhɑ郁bV

	rxTexObj2D m_texDepth;			//!< fvX}bv`peNX`


public:
	int m_iDraw;			//!< `tO
	int m_iPolyDraw;		//!< |S`tO
	int m_iDepthDraw;		//!< fvX}bv`tO

	int m_iViewMesh;		//!< _ύXăbVtO
	int m_iFiltering;	//!< fvX}bvtB^O
	int m_iUseGPU;			//!< GPUgptO

public:
	//! RXgN^
	rxFlGLWindow(int x, int y, int w, int h, const char* l, void *parent);

	//! fXgN^
	~rxFlGLWindow();

public:
	void InitGL(void);
	void Resize(int w, int h);

	void Projection(int w, int h);
	void InitView(void);
	vector<string> SetDrawString(void);


	void Display(void);
	void Mouse(int button, int state, int x, int y);
	void Motion(int x, int y);
	void PassiveMotion(int x, int y);
	void Idle(void);
	void Timer(void);
	void Keyboard(int key, int x, int y);

	static void OnTimer_s(void* x);
	static void OnIdle_s(void* x);

	void SwitchIdle(int on);
	void SwitchFullScreen(void);
	
	void SaveDepthMap(const string fn);
	void SaveDisplay(const string &fn);

	// SSM
	void InitParticles(void);
	void CreateSSM(int debug_output = 1);
	void SelectGrid(int x, int y);
	void OutputGridInfo(int grid);
	void DrawPolygons(rxPolygons &polys, int draw);

	void RenderPolygon(int w, int h);
	void RenderDepthMap(int w, int h);


private:
	//! `R[obN֐̃I[o[Ch
	void draw(void)
	{
		if(!context_valid()) InitGL();
		if(!valid()) Resize(w(), h());
		Display();    // OpenGL`
    }

	//! TCYR[obN֐̃I[o[Ch
	void resize(int x_, int y_, int w_, int h_)
	{
		Fl_Gl_Window::resize(x_, y_, w_, h_);
		//Resize(w_, h_);
    }

public:
	int handle(int e);	// handle֐̃I[o[Ch

	// j[Cxgnh
	static void OnMenuDraw_s(Fl_Widget *widget, void* x);
	inline void OnMenuDraw(double val, string label);
	static void OnMenuDrawDepthMap_s(Fl_Widget *widget, void* x);
	inline void OnMenuDrawDepthMap(double val, string label);
	static void OnMenuDrawPolygon_s(Fl_Widget *widget, void* x);
	inline void OnMenuDrawPolygon(double val, string label);

	static void OnMenuSSM_s(Fl_Widget *widget, void* x);
	inline void OnMenuSSM(double val, string label);

	static void OnSpinSSM_s(Fl_Widget *widget, void* x);
	inline void OnSpinSSM(double val, string label);
};




#endif // #ifdef _RX_FLTK_GLCANVAS_H_
