/*! 
  @file rx_ssm.cu
	
  @brief SSM@ɂ郁bV
 		- M. Muller et al., Screen space meshes, SCA2007, 2007. 

  @author Makoto Fujisawa
  @date 2011-05
*/
// FILE --rx_ssm.cu--


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cutil_inline.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>

#include <iostream>
#include <fstream>

#include <GL/glew.h>

#include <cutil.h>
#include <cuda_gl_interop.h>

#include <GL/freeglut.h>

#include "rx_ssm_kernel.cu"
#include "rx_ssm_tables.h"
#include "cudpp.h"



//-----------------------------------------------------------------------------
// MARK:O[oϐ
//-----------------------------------------------------------------------------
CUDPPHandle g_cudppScanPlanSSM;	//!< Prefix Sum(Scan)vZ

// MS@̃e[u
uint* g_puSSMeshTable = 0;
uint* g_puSSEdgeTable = 0;
uint* g_puSSNodeTable = 0;
uint* g_puSSVRotTable = 0;

//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
void CuAllocateArray(void **devPtr, size_t size);
void CuFreeArray(void *devPtr);


//Round a / b to nearest higher integer value
uint iDivUp(uint a, uint b)
{
	return (a % b != 0) ? (a / b + 1) : (a / b);
}

// compute grid and thread block size for a given number of elements
void computeGridSize(uint n, uint blockSize, uint &numBlocks, uint &numThreads)
{
	numThreads = min(blockSize, n);
	numBlocks = iDivUp(n, numThreads);
}


/*!
 * Scanv쐬
 */
void CuCreateScanPlan(int num)
{
	CUDPPConfiguration config;
	config.algorithm = CUDPP_SCAN;
	config.datatype = CUDPP_UINT;
	config.op = CUDPP_ADD;
	config.options = CUDPP_OPTION_FORWARD | CUDPP_OPTION_EXCLUSIVE;
	cudppPlan(&g_cudppScanPlanSSM, config, num, 1, 0);
}


/*!
 * _̏
 * @param[in] pack
 */
void CuInitVPackf(rxVPackf &pack, int size)
{
	if(pack.dPos) CuFreeArray(pack.dPos);
	CuAllocateArray((void**)&pack.dPos, sizeof(float)*3*size);
	if(pack.dOcc) CuFreeArray(pack.dOcc);
	CuAllocateArray((void**)&pack.dOcc, sizeof(uint)*size);
	if(pack.dOccScan) CuFreeArray(pack.dOccScan);
	CuAllocateArray((void**)&pack.dOccScan, sizeof(uint)*size);
}

/*!
 * _̏
 * @param[in] pack
 */
void CuInitVPacke(rxVPacke &pack, int size)
{
	if(pack.dPos) CuFreeArray(pack.dPos);
	CuAllocateArray((void**)&pack.dPos, sizeof(rxSSEdgeG)*3*size);
	if(pack.dOcc) CuFreeArray(pack.dOcc);
	CuAllocateArray((void**)&pack.dOcc, sizeof(uint)*size);
	if(pack.dOccScan) CuFreeArray(pack.dOccScan);
	CuAllocateArray((void**)&pack.dOccScan, sizeof(uint)*size);
}

/*!
 * _̏
 * @param[in] pack
 */
void CuCleanVPackf(rxVPackf &pack)
{
	if(pack.dPos) CuFreeArray(pack.dPos);
	if(pack.dCompactedPos) CuFreeArray(pack.dCompactedPos);
	if(pack.dOcc) CuFreeArray(pack.dOcc);
	if(pack.dOccScan) CuFreeArray(pack.dOccScan);
}

/*!
 * _̏
 * @param[in] pack
 */
void CuCleanVPacke(rxVPacke &pack)
{
	if(pack.dPos) CuFreeArray(pack.dPos);
	if(pack.dCompactedPos) CuFreeArray(pack.dCompactedPos);
	if(pack.dOcc) CuFreeArray(pack.dOcc);
	if(pack.dOccScan) CuFreeArray(pack.dOccScan);
}

/*!
 * MS@̃e[u
 */
void CuInitTable(void)
{
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindSigned);

	cutilSafeCall(cudaMalloc((void**) &g_puSSMeshTable, 256*19*sizeof(int)));
	cutilSafeCall(cudaMemcpy(g_puSSMeshTable, g_MeshTable, 256*19*sizeof(int), cudaMemcpyHostToDevice) );
	cutilSafeCall(cudaBindTexture(0, g_texSSMeshTable, g_puSSMeshTable, channelDesc) );

	cutilSafeCall(cudaMalloc((void**) &g_puSSEdgeTable, 16*4*sizeof(int)));
	cutilSafeCall(cudaMemcpy(g_puSSEdgeTable, g_EdgeTable, 16*4*sizeof(int), cudaMemcpyHostToDevice) );
	cutilSafeCall(cudaBindTexture(0, g_texSSEdgeTable, g_puSSEdgeTable, channelDesc) );

	cutilSafeCall(cudaMalloc((void**) &g_puSSNodeTable, 4*8*sizeof(int)));
	cutilSafeCall(cudaMemcpy(g_puSSNodeTable, g_NodeTable, 4*8*sizeof(int), cudaMemcpyHostToDevice) );
	cutilSafeCall(cudaBindTexture(0, g_texSSNodeTable, g_puSSNodeTable, channelDesc) );

	cutilSafeCall(cudaMalloc((void**) &g_puSSVRotTable, 4*14*sizeof(int)));
	cutilSafeCall(cudaMemcpy(g_puSSVRotTable, g_VrtRotTable, 4*14*sizeof(int), cudaMemcpyHostToDevice) );
	cutilSafeCall(cudaBindTexture(0, g_texSSVRotTable, g_puSSVRotTable, channelDesc) );
}

/*!
 * MS@̃e[u̔j
 */
void CuCleanTable(void)
{
	cutilSafeCall(cudaFree(g_puSSMeshTable));
	cutilSafeCall(cudaFree(g_puSSEdgeTable));
	cutilSafeCall(cudaFree(g_puSSNodeTable));
	cutilSafeCall(cudaFree(g_puSSVRotTable));
}


//-----------------------------------------------------------------------------
// MARK:SSM
//-----------------------------------------------------------------------------

/*!
 * p[^RX^gɓ]
 */
void CuSetParameters(rxSsmParams *hostParams)
{
	cutilSafeCall( cudaMemcpyToSymbol(g_cSSMParams, hostParams, sizeof(rxSsmParams)) );
}

/*!
 * fvX}bvŏ
 * @param[in] dDMap fvX}bv((nx+1)x(ny+1))
 * @param[in] nx,ny Obh𑜓x
 */
void CuInitDepthMap(float* dDMap, int nx, int ny)
{
	int n = (nx+1)*(ny+1);

	// 1Xbh/sNZ
	uint numThreads, numBlocks;
	computeGridSize(n, 64, numBlocks, numThreads);

	// J[ls
	initDepthMap<<< numBlocks, numThreads >>>(dDMap, n);

	cutilCheckMsg("initDepthMap kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNfvX}bv̌vZ
 * @param[in] dDMap fvX}bv((nx+1)x(ny+1)xfloat)
 * @param[inout] dPrtPos p[eBNW(pnumxfloat3)CXN[Xy[Xł̍WԂ
 * @param[out] dPrtRad XN[Xy[Xł̃p[eBNa(pnumxfloat3)
 * @param[in] pnum p[eBN
 * @param[in] tr p[eBNavZpW(vf3)
 * @param[in] pmv eƃfr[s|s(vf16=4x4)
 * @param[in] W,H XN[̉𑜓x
 * @param[in] dw,dh Obh
 * @param[in] nx,ny Obh𑜓x
 */
void CuCreateDepthMap(float* dDMap, float* dPrtPos, float* dPrtRad, int pnum, float* tr, float* pmv, 
					  int W, int H, float dw, float dh, int nx, int ny)
{
	float3 tr3;
	tr3.x = tr[0];
	tr3.y = tr[1];
	tr3.z = tr[2];

	matrix4x4 mat_pmv;
	for(int i = 0; i < 4; ++i){
		mat_pmv.e[i].x = pmv[4*i+0];
		mat_pmv.e[i].y = pmv[4*i+1];
		mat_pmv.e[i].z = pmv[4*i+2];
		mat_pmv.e[i].w = pmv[4*i+3];
	}

	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(pnum, 64, numBlocks, numThreads);

	// J[ls
	calDepthMap<<< numBlocks, numThreads >>>(dDMap, (float3*)dPrtPos, (float3*)dPrtRad, pnum, tr3, mat_pmv, 
											 (float)W, (float)H, dw, dh, nx, ny);

	cutilCheckMsg("calDepthMap kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * fvX}bvɕ{
 * @param[in] dDMap fvX}bv((nx)x(ny))
 * @param[in] nx,ny }bv𑜓x
 * @param[in] n_filter tB^
 * @param[in] binomials 񍀉ZW(n_filter = 0, 1, 2,..,RX_MAX_FILTER_SIZE ̌WԂɊi[Ă)
 * @param[in] zmax ֊sƂȂfvX臒l
 */
void CuDepthSmoothing(float* dDMap, int nx, int ny, int n_filter, float *binomials, float zmax)
{
	int b = (n_filter+1)*(n_filter+1);
	cudaMemcpyToSymbol(g_cBinomials, binomials, sizeof(float)*b);

	// 1Xbh/sNZ(eXbh}bṽsNZɑΉ悤2Iɔzu)
	dim3 block(BLOCK_SIZE, BLOCK_SIZE);
    dim3 grid((nx+block.x-1)/block.x, (ny+block.y-1)/block.y);

	// ubNƂ̃VFA[hTCY(ő16KB)
	int sbytes = (block.x+2*n_filter)*(block.y+2*n_filter)*sizeof(float);

	// J[ls(x)
	smoothDepthMapX<<< grid, block, sbytes >>>(dDMap, nx, ny, n_filter, block.x+2*n_filter, zmax);

	cutilCheckMsg("smoothDepthMapX kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	// J[ls(y)
	smoothDepthMapY<<< grid, block, sbytes >>>(dDMap, nx, ny, n_filter, block.x+2*n_filter, zmax);

	cutilCheckMsg("smoothDepthMapY kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * ֊sGbW̌ofront edge vertex̌vZ
 * @param[in] dDMap fvX}bv(TCY = (nx+1)x(ny+1))
 * @param[in] nx,ny }bv𑜓x
 * @param[in] dw,dh Obh
 * @param[in] zmax ֊sƂȂfvX臒l
 * @param[out] dNodeVrts m[h_(TCY = (nx+1)x(ny+1))
 * @param[out] dMGrid bVpObh(TCY = (nx)x(ny))
 * @param[out] num_node_vertex m[h_
 */
void CuCalNodeVertex(float* dDMap, int nx, int ny, float dw, float dh, float zmax, 
					 rxVPackf &dNodeVrts, rxSSGridG* dMGrid, int &num_node_vertex)
{
	uint block1, grid1;
	dim3 block2, grid2;
	unsigned int size = (nx+1)*(ny+1);
	uint last_val, last_scan_val;

	CuCreateScanPlan(size);

	cutilSafeCall(cudaMemset((void*)dNodeVrts.dOcc, 0, size*sizeof(uint)));

	//
	// m[h_
	//
	// 1Xbh/m[h
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx+1)+block2.x-1)/block2.x, ((ny+1)+block2.y-1)/block2.y);
	
	// J[ls
	calNodeVertex<<< grid2, block2 >>>(dDMap, nx+1, ny+1, dw, dh, zmax, (float3*)dNodeVrts.dPos, dNodeVrts.dOcc);

	cutilCheckMsg("calNodeVertex kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂


	//
	// m[h_̃pbLO
	//
	// m[h_Ȃ1,łȂȂ0i[ꂽzScan
	cudppScan(g_cudppScanPlanSSM, dNodeVrts.dOccScan, dNodeVrts.dOcc, size);

	// Exclusive scan (Ō̗vf0Ԗڂn-2Ԗڂ܂ł̍vɂȂĂ)Ȃ̂ŁC
	// ScanOz̍Ō(n-1Ԗ)̗vfƍv邱Ƃŗ֊sGbWvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dNodeVrts.dOcc+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void *) &last_scan_val, (void*)(dNodeVrts.dOccScan+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_node_vertex = last_val+last_scan_val;

	// l߂֊sGbWi[̈̊m
	if(dNodeVrts.dCompactedPos) cutilSafeCall(cudaFree(dNodeVrts.dCompactedPos));
	cutilSafeCall(cudaMalloc((void**)&dNodeVrts.dCompactedPos, num_node_vertex*3*sizeof(float)));

	// ֊sGbWl߂
	computeGridSize(size, 64, grid1, block1);

	// J[ls
	compactNodeVertex<<<grid1, block1>>>((float3*)(dNodeVrts.dCompactedPos), dNodeVrts.dOcc, dNodeVrts.dOccScan, (float3*)(dNodeVrts.dPos), size);
	
	cutilCheckMsg("compactNodeVertex kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂


	//
	// m[h_CfbNXbVObhɊi[
	//
	// 1Xbh/Obh
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);
		
	// J[ls
	storeNodeVertex<<< grid2, block2 >>>(dMGrid, nx, ny, (float3*)(dNodeVrts.dCompactedPos), dNodeVrts.dOcc, dNodeVrts.dOccScan);

	cutilCheckMsg("storeNodeVertex kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	cudppDestroyPlan(g_cudppScanPlanSSM);
}



/*!
 * ֊sGbW̌ofront edge vertex̌vZ
 * @param[in] dDMap fvX}bv((nx+1)x(ny+1))
 * @param[in] nx,ny }bv𑜓x
 * @param[in] dw,dh Obh
 * @param[in] zmax ֊sƂȂfvX臒l
 * @param[in] dPrtPos p[eBNW(pnumxfloat3)CXN[Xy[Xł̍WԂ
 * @param[in] dPrtRad XN[Xy[Xł̃p[eBNa(pnumxfloat3)
 * @param[in] pnum p[eBN
 * @param[in] W,H XN[̉𑜓x
 * @param[out] dEdge GbW(nx*(ny+1)+(nx+1)*ny)
 * @param[out] dCompactedEdge ֊sGbW(GbW񂩂֊sGbŴ݂l߂)
 * @param[out] dEdgeSil ֊sGbWȂ1,łȂ0i[z(nx*(ny+1)+(nx+1)*ny)
 * @param[out] dEdgeSilScan dEdgeSilPrefix sum (nx*(ny+1)+(nx+1)*ny)
 * @param[out] dEdgeVrts GbW_(TCY = (nx*(ny+1)+(nx+1)*ny))
 * @param[out] dBackVrts wʃGbW_(TCY = (nx*(ny+1)+(nx+1)*ny))
 * @param[out] dMGrid bVpObh(TCY = (nx)x(ny))
 * @param[in] num_nv m[h_
 * @param[out] num_edge ֊sGbW
 * @param[out] num_front_edge GbW_(O)
 * @param[out] num_back_edge GbW_(w)
 */
void CuDetectSilhouetteEdgeVertex(float* dDMap, int nx, int ny, float dw, float dh, float zmax, 
								  float* dPrtPos, float* dPrtRad, int pnum, int W, int H, 
								  rxVPacke &dEdge, rxVPackf &dEdgeVrts, rxVPackf &dBackVrts, rxSSGridG* dMGrid, 
								  int num_nv, int &num_edge, int &num_front_edge, int &num_back_edge)
{
	uint block1, grid1;
	dim3 block2, grid2;
	unsigned int size = nx*(ny+1)+(nx+1)*ny;
	int offset = (nx)*(ny+1);
	uint last_val, last_scan_val;

	CuCreateScanPlan(size);

	//
	// ֊sGbWo
	//
	cutilSafeCall(cudaMemset((void*)dEdge.dOcc, 0, size*sizeof(uint)));

	// x֊sGbWo
	// 1Xbh/GbW
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny+1)+block2.y-1)/block2.y);
	
	// J[ls
	detectSilhouetteEdgeX<<< grid2, block2 >>>(dDMap, nx, ny+1, dw, dh, zmax, dEdge.dPos, dEdge.dOcc);

	cutilCheckMsg("detectSilhouetteEdgeX kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	
	// y֊sGbWo
	// 1Xbh/GbW
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx+1)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);
		
	// J[ls
	detectSilhouetteEdgeY<<< grid2, block2 >>>(dDMap, nx+1, ny, dw, dh, zmax, dEdge.dPos+offset, dEdge.dOcc+offset);

	cutilCheckMsg("detectSilhouetteEdgeY kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	
	//
	// ֊sGbW̉ۂPrefix Sum쐬āCGbWl߂
	//
	// ֊sGbWȂ1,łȂȂ0i[ꂽzScan
	cudppScan(g_cudppScanPlanSSM, dEdge.dOccScan, dEdge.dOcc, size);

	// Exclusive scan (Ō̗vf0Ԗڂn-2Ԗڂ܂ł̍vɂȂĂ)Ȃ̂ŁC
	// ScanOz̍Ō(n-1Ԗ)̗vfƍv邱Ƃŗ֊sGbWvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dEdge.dOcc+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void*)&last_scan_val, (void*)(dEdge.dOccScan+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_edge = last_val+last_scan_val;

	// l߂֊sGbWi[̈̊m
	if(dEdge.dCompactedPos) cutilSafeCall(cudaFree(dEdge.dCompactedPos));
	cutilSafeCall(cudaMalloc((void**)&dEdge.dCompactedPos, num_edge*sizeof(rxSSEdgeG)));

	// ֊sGbWl߂
	computeGridSize(size, 64, grid1, block1);

	// J[ls
	compactSilhouetteEdges<<<grid1, block1>>>(dEdge.dCompactedPos, dEdge.dOcc, dEdge.dOccScan, dEdge.dPos, size);
	
	cutilCheckMsg("CompactSilhouetteEdges kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂



	//
	// ֊sGbWfront vertexZo
	//
	// GbW_Lz̏
	cutilSafeCall(cudaMemset((void*)dEdgeVrts.dOcc, 0, size*sizeof(uint)));

	// 1Xbh/p[eBN
	computeGridSize(pnum, 256, grid1, block1);

	// J[ls
	calFrontEdgeVertex<<<grid1, block1>>>((float3*)dPrtPos, (float3*)dPrtRad, pnum, 
										  dEdge.dPos, dEdge.dOcc, offset, nx, ny, dw, dh, W, H, 
										  (float3*)dEdgeVrts.dPos, dEdgeVrts.dOcc);

	cutilCheckMsg("calFrontEdgeVertex kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	
	//
	// GbW_LPrefix Sum쐬āCGbW_l߂
	//
	// ֊sGbWȂ1,łȂȂ0i[ꂽzScan
	cudppScan(g_cudppScanPlanSSM, dEdgeVrts.dOccScan, dEdgeVrts.dOcc, size);

	// ֊sGbWvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dEdgeVrts.dOcc+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void*)&last_scan_val, (void*)(dEdgeVrts.dOccScan+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_front_edge = last_val+last_scan_val;

	// l߂֊sGbWi[̈̊m
	if(dEdgeVrts.dCompactedPos) cutilSafeCall(cudaFree(dEdgeVrts.dCompactedPos));
	cutilSafeCall(cudaMalloc((void**)&dEdgeVrts.dCompactedPos, num_front_edge*sizeof(float)*3));

	// ֊sGbWl߂
	computeGridSize(size, 64, grid1, block1);

	// J[ls
	compactEdgeVertices<<<grid1, block1>>>((float3*)dEdgeVrts.dCompactedPos, dEdgeVrts.dOcc, dEdgeVrts.dOccScan, (float3*)dEdgeVrts.dPos, dEdge.dPos, size);
	
	cutilCheckMsg("compactEdgeVertices kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂



	//
	// ֊sGbWback vertexZo
	//
	// GbW_Lz̏
	cutilSafeCall(cudaMemset((void*)dBackVrts.dOcc, 0, size*sizeof(uint)));

	// xGbW
	// 1Xbh/GbW
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny+1)+block2.y-1)/block2.y);
	
	// J[ls
	calBackEdgeVertexX<<<grid2, block2>>>(dDMap, dEdge.dPos, dEdge.dOcc, nx, ny+1, dw, dh, W, H, 
										  (float3*)dEdgeVrts.dPos, (float3*)dBackVrts.dPos, dBackVrts.dOcc);

	cutilCheckMsg("calBackEdgeVertexX kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	offset = (nx)*(ny+1);

	// yGbW
	// 1Xbh/GbW
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx+1)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);
	
	// J[ls
	calBackEdgeVertexY<<<grid2, block2>>>(dDMap, dEdge.dPos, dEdge.dOcc, nx+1, ny, dw, dh, W, H, 
										  (float3*)dEdgeVrts.dPos, (float3*)dBackVrts.dPos, dBackVrts.dOcc, offset);

	cutilCheckMsg("calBackEdgeVertexY kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	//
	// GbW_LPrefix Sum쐬āCGbW_l߂
	//
	// ֊sGbWȂ1,łȂȂ0i[ꂽzScan
	cudppScan(g_cudppScanPlanSSM, dBackVrts.dOccScan, dBackVrts.dOcc, size);

	// ֊sGbWvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dBackVrts.dOcc+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void*)&last_scan_val, (void*)(dBackVrts.dOccScan+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_back_edge = last_val+last_scan_val;

	if(num_back_edge){
		// l߂֊sGbWi[̈̊m
		if(dBackVrts.dCompactedPos) cutilSafeCall(cudaFree(dBackVrts.dCompactedPos));
		cutilSafeCall(cudaMalloc((void**)&dBackVrts.dCompactedPos, num_back_edge*sizeof(float)*3));

		// ֊sGbWl߂
		computeGridSize(size, 64, grid1, block1);

		// J[ls
		compactBackEdgeVertices<<<grid1, block1>>>((float3*)dBackVrts.dCompactedPos, dBackVrts.dOcc, dBackVrts.dOccScan, (float3*)dBackVrts.dPos, size);
	
		cutilCheckMsg("compactEdgeVertices kernel execution failed");	// J[lsG[`FbN
		cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	}


	//
	// GbW_CfbNXbVObhɊi[
	//
	// 1Xbh/Obh
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);
		
	// J[ls
	storeEdgeVertex<<< grid2, block2 >>>(dMGrid, nx, ny, offset, num_nv, num_nv+num_front_edge, 
										 (float3*)(dEdgeVrts.dCompactedPos), dEdgeVrts.dOcc, dEdgeVrts.dOccScan, 
										 (float3*)(dBackVrts.dCompactedPos), dBackVrts.dOcc, dBackVrts.dOccScan);

	cutilCheckMsg("storeEdgeVertex kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	cudppDestroyPlan(g_cudppScanPlanSSM);
}



/*!
 * bV
 * @param[in] dDMap fvX}bv(TCY = (nx+1)x(ny+1))
 * @param[in] nx,ny }bv𑜓x
 * @param[in] dw,dh Obh
 * @param[in] zmax ֊sƂȂfvX臒l
 * @param[out] dTriNum eObh̃bV
 * @param[out] dTriNumScan eObh̃bVScan
 * @param[out] dBack2Vrts ŔwʃGbW_(TCY = (nx*(ny+1)+(nx+1)*ny))
 * @param[in] dVrts bV_
 * @param[in] num_vrts bV_
 * @param[out] num_back2_vrts ŔwʃGbW_
 * @param[out] dTriArray bV
 * @param[out] num_mesh bV
 */
void CuCalMesh(rxSSGridG* dMGrid, int nx, int ny, float dw, float dh, float zmax, uint* dTriNum, uint* dTriNumScan, 
			   rxVPackf &dBack2Vrts, float* dVrts, int num_vrts, int &num_back2_vrts, uint* &dTriArray, int &num_mesh)
{
	uint block1, grid1;
	dim3 block2, grid2;
	unsigned int size = nx*ny;
	unsigned int size_e = nx*(ny+1)+(nx+1)*ny;
	int offset = (nx)*(ny+1);
	uint last_val, last_scan_val;


	// GbW_Lz̏
	cutilSafeCall(cudaMemset((void*)dBack2Vrts.dOcc, 0, size_e*sizeof(uint)));

	// bVz̏
	cutilSafeCall(cudaMemset((void*)dTriNum, 0, size*sizeof(uint)));

	CuCreateScanPlan(size_e);
	//
	// eObhŃbV
	//
	// 1Xbh/Obh
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);

	// J[ls
	calGridMesh<<< grid2, block2 >>>(dMGrid, nx, ny, zmax, (float3*)dVrts, num_vrts, dTriNum, 
									 (float3*)dBack2Vrts.dPos, dBack2Vrts.dOcc, offset);

	cutilCheckMsg("calGridMesh kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	
	//
	// GbW_LPrefix Sum쐬āCGbW_l߂
	//
	// GbW_Ȃ1,łȂȂ0i[ꂽzScan
	cudppScan(g_cudppScanPlanSSM, dBack2Vrts.dOccScan, dBack2Vrts.dOcc, size_e);

	// ֊sGbWvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dBack2Vrts.dOcc+size_e-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void*)&last_scan_val, (void*)(dBack2Vrts.dOccScan+size_e-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_back2_vrts = last_val+last_scan_val;

	if(num_back2_vrts){
		// l߂֊sGbWi[̈̊m
		if(dBack2Vrts.dCompactedPos) cutilSafeCall(cudaFree(dBack2Vrts.dCompactedPos));
		cutilSafeCall(cudaMalloc((void**)&dBack2Vrts.dCompactedPos, num_back2_vrts*sizeof(float)*3));

		// ֊sGbWl߂
		computeGridSize(size_e, 256, grid1, block1);

		// J[ls
		compactBackEdgeVertices<<<grid1, block1>>>((float3*)dBack2Vrts.dCompactedPos, dBack2Vrts.dOcc, dBack2Vrts.dOccScan, (float3*)dBack2Vrts.dPos, size_e);
	
		cutilCheckMsg("compactBackEdgeVertices kernel execution failed");	// J[lsG[`FbN
		cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	}

	cudppDestroyPlan(g_cudppScanPlanSSM);


	CuCreateScanPlan(size);

	// bVzScan
	cudppScan(g_cudppScanPlanSSM, dTriNumScan, dTriNum, size);
	
	// bVvZ
	cutilSafeCall(cudaMemcpy((void*)&last_val, (void*)(dTriNum+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy((void*)&last_scan_val, (void*)(dTriNumScan+size-1), sizeof(uint), cudaMemcpyDeviceToHost));
	num_mesh = last_val+last_scan_val;

	// bVi[̈̊m
	if(dTriArray) cutilSafeCall(cudaFree(dTriArray));
	cutilSafeCall(cudaMalloc((void**)&dTriArray, num_mesh*sizeof(uint)*3));

	// 1Xbh/Obh
	block2 = dim3(BLOCK_SIZE, BLOCK_SIZE);
    grid2 = dim3(((nx)+block2.x-1)/block2.x, ((ny)+block2.y-1)/block2.y);

	//printf("num_vrts = %d\n", num_vrts);

	// J[ls
	genGridMesh<<< grid2, block2 >>>(dMGrid, nx, ny, zmax, (float3*)dVrts, num_vrts, dTriArray, dTriNumScan, 
									 (float3*)dBack2Vrts.dPos, dBack2Vrts.dOcc, dBack2Vrts.dOccScan, offset, num_vrts);

	cutilCheckMsg("genGridMesh kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

	cudppDestroyPlan(g_cudppScanPlanSSM);
}

/*!
 * ֊s
 * @param[inout] dVrts bV_
 * @param[in] num_vrts bV_
 * @param[in] num_node_vrts m[h_
 * @param[in] dTriArray bV
 * @param[in] num_mesh bV
 * @param[in] n_iter 
 */
void CuSilhouetteSmoothing(float* dVrts, int num_vrts, int num_node_vrts, uint* dTriArray, int num_mesh, int n_iter)
{
	uint block1, grid1;

	float4 *dAvgPos;
	cutilSafeCall(cudaMalloc(&dAvgPos, (num_vrts-num_node_vrts)*sizeof(float4)));


	for(int l = 0; l < n_iter; ++l){
		cutilSafeCall(cudaMemset(dAvgPos, 0, (num_vrts-num_node_vrts)*sizeof(float4)));

		// 1Xbh/bV
		computeGridSize(num_mesh, 256, grid1, block1);

		// J[ls(ψʒu̎Zo)
		smoothSilhouette<<< grid1, block1 >>>((float3*)dVrts, num_node_vrts, dTriArray, num_mesh, dAvgPos);

		cutilCheckMsg("smoothSilhouette kernel execution failed");	// J[lsG[`FbN
		cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂


		// 1Xbh/_
		computeGridSize(num_vrts-num_node_vrts, 256, grid1, block1);

		// J[ls(ψʒu̎Zo)
		smoothSilhouette2<<< grid1, block1 >>>((float3*)dVrts, num_vrts, num_node_vrts, dAvgPos);

		cutilCheckMsg("smoothSilhouette2 kernel execution failed");	// J[lsG[`FbN
		cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
	}
}

/*!
 * XN[Xy[X璸_3DԂɖ߂
 * @param[in] dSSVrts XN[Xy[XbV_
 * @param[in] num_vrts bV_
 * @param[in] mvq eƃfr[ts|s(vf16=4x4)
 * @param[in] W,H XN[̉𑜓x
 * @param[out] dVrts 3DbV_
 */
void CuTransformBack(float* dSSVrts, int num_vrts, float* mvq, float* q, int W, int H, float* dVrts)
{
	uint block1, grid1;

	float4 vec_q;
	vec_q.x = q[0];
	vec_q.y = q[1];
	vec_q.z = q[2];
	vec_q.w = q[3];

	matrix4x4 mat_mvq;
	for(int i = 0; i < 4; ++i){
		mat_mvq.e[i].x = mvq[4*i+0];
		mat_mvq.e[i].y = mvq[4*i+1];
		mat_mvq.e[i].z = mvq[4*i+2];
		mat_mvq.e[i].w = mvq[4*i+3];
	}

	// 1Xbh/bV
	computeGridSize(num_vrts, 256, grid1, block1);

	// J[ls(ψʒu̎Zo)
	transfromBack<<< grid1, block1 >>>((float3*)dSSVrts, (float3*)dVrts, num_vrts, mat_mvq, vec_q, W, H);

	cutilCheckMsg("transfromBack kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂

}

/*!
 * _@vZ
 * @param[in] dVrts bV_
 * @param[in] num_vrts bV_
 * @param[in] dTriArray bV
 * @param[in] num_mesh bV
 * @param[out] dNrms _@
 */
void CuCalVertexNormal(float* dVrts, int num_vrts, uint* dTriArray, int num_mesh, float* dNrms)
{
	uint block1, grid1;

	// @z̏
	cutilSafeCall(cudaMemset(dNrms, 0, num_vrts*sizeof(float3)));

	// 1Xbh/bV
	computeGridSize(num_mesh, 256, grid1, block1);

	// J[ls(ψʒu̎Zo)
	sumFaceNormal<<< grid1, block1 >>>((float3*)dVrts, num_vrts, dTriArray, num_mesh, dNrms);

	cutilCheckMsg("sumFaceNormal kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂


	// 1Xbh/_
	computeGridSize(num_vrts, 256, grid1, block1);

	// J[ls(ψʒu̎Zo)
	normalizeNormal<<< grid1, block1 >>>((float3*)dNrms, num_vrts);

	cutilCheckMsg("normalizeNormal kernel execution failed");	// J[lsG[`FbN
	cutilSafeCall(cudaThreadSynchronize());		// SẴXbhÎ҂
}


//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
/*!
 * CUDAfoCX̐ݒ
 *  - R}hCɊÂCUDAfoCXݒ(()-device 0)
 * @param[in] argc R}hC̐
 * @param[in] argv R}hCXg(argv[0]͎st@C)
 */
void CuInit(int argc, char **argv)
{   
	if(cutCheckCmdLineFlag(argc, (const char**)argv, "device")){
		cutilDeviceInit(argc, argv);
	}
	else{
		cudaSetDevice( cutGetMaxGflopsDeviceId() );
	}
}

/*!
 * CUDAfoCX̐ݒ - id𒼐ڎw
 * @param[in] id foCXID
 */
void CuSetDevice(int id)
{ 
	int device_count = 0;
	cudaGetDeviceCount(&device_count);
	if(id < 0 || id >= device_count){
		id = cutGetMaxGflopsDeviceId();
		cudaSetDevice(id);
	}
	else{
		cudaSetDevice(id);
	}
}

/*!
 * foCX̊m
 * @param[out] dPtr foCXւ̃|C^
 * @param[in] size mۃTCY(̃TCY)
 */
void CuAllocateArray(void **dPtr, size_t size)
{
	cutilSafeCall(cudaMalloc(dPtr, size));
}

/*!
 * foCX̉
 * @param[in] devPtr foCXւ̃|C^
 */
void CuFreeArray(void *dPtr)
{
	cutilSafeCall(cudaFree(dPtr));
}

/*!
 * foCẌ̏
 * @param[in] dPtr foCXւ̃|C^
 * @param[in] val l
 * @param[in] size ̈̃TCY(̃TCY)
 */
void CuSetArrayValue(void *dPtr, int val, size_t size)
{
	cutilSafeCall(cudaMemset(dPtr, val, size));
}

/*!
 * foCXԃRs[
 * @param[in] dDst Rs[
 * @param[in] dSrc Rs[
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayD2D(void *dDst, void *dSrc, int size)
{
	cutilSafeCall(cudaMemcpy(dDst, dSrc, size, cudaMemcpyDeviceToDevice));
}

/*!
 * foCXzXgւ̃Rs[
 * @param[in] hDst Rs[zXg(ŒsizemۂĂ邱)
 * @param[in] dSrc Rs[foCX
 * @param[in] vbo dSrcVBȌꍇCVBOIDDłȂꍇ0w
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayFromDevice(void* hDst, const void* dSrc, unsigned int vbo, int size)
{   
	if(vbo) cutilSafeCall(cudaGLMapBufferObject((void**)&dSrc, vbo));

	cutilSafeCall(cudaMemcpy(hDst, dSrc, size, cudaMemcpyDeviceToHost));
	
	if(vbo) cutilSafeCall(cudaGLUnmapBufferObject(vbo));
}

/*!
 * zXgfoCXւ̃Rs[
 * @param[in] dDst Rs[foCX(ŒsizemۂĂ邱)
 * @param[in] hSrc Rs[zXg
 * @param[in] offset Rs[ItZbg
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayToDevice(void* dDst, const void* hSrc, int offset, int size)
{
	cutilSafeCall(cudaMemcpy((char*)dDst+offset, hSrc, size, cudaMemcpyHostToDevice));
}

/*!
 * VBO}bsO
 * @param[in] vbo VBO,PBO
 */
void *CuMapGLBufferObject(uint vbo)
{
	void *ptr;
	cutilSafeCall(cudaGLMapBufferObject(&ptr, vbo));
	return ptr;
}

/*!
 * VBOA}bv
 * @param[in] vbo VBO,PBO
 */
void CuUnmapGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLUnmapBufferObject(vbo));
}

/*!
 * PBO,VBOobt@CUDAɓo^
 * @param[in] vbo VBO,PBO
 */
void CuRegisterGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLRegisterBufferObject(vbo));
}

/*!
 * PBO,VBOobt@CUDA폜
 * @param[in] vbo VBO,PBO
 */
void CuUnregisterGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLUnregisterBufferObject(vbo));
}

/*!
 * Xbh
 */
void CuThreadSync(void)
{
	cutilSafeCall(cudaThreadSynchronize());
}

/*!
 * foCXvpeB̕\
 */
void CuDeviceProp(void)
{
	int n;	//foCX
	cutilSafeCall(cudaGetDeviceCount(&n));

	for(int i = 0; i < n; ++i){
		cudaDeviceProp dev;

		// foCXvpeB擾
		cutilSafeCall(cudaGetDeviceProperties(&dev, i));

		printf("device %d\n", i);
		printf(" device name : %s\n", dev.name);
		printf(" total global memory : %d (MB)\n", dev.totalGlobalMem/1024/1024);
		printf(" shared memory / block : %d (KB)\n", dev.sharedMemPerBlock/1024);
		printf(" register / block : %d\n", dev.regsPerBlock);
		printf(" warp size : %d\n", dev.warpSize);
		printf(" max pitch : %d (B)\n", dev.memPitch);
		printf(" max threads / block : %d\n", dev.maxThreadsPerBlock);
		printf(" max size of each dim. of block : (%d, %d, %d)\n", dev.maxThreadsDim[0], dev.maxThreadsDim[1], dev.maxThreadsDim[2]);
		printf(" max size of each dim. of grid  : (%d, %d, %d)\n", dev.maxGridSize[0], dev.maxGridSize[1], dev.maxGridSize[2]);
		printf(" clock rate : %d (MHz)\n", dev.clockRate/1000);
		printf(" total constant memory : %d (KB)\n", dev.totalConstMem/1024);
		printf(" compute capability : %d.%d\n", dev.major, dev.minor);
		printf(" alignment requirement for texture : %d\n", dev.textureAlignment);
		printf(" device overlap : %s\n", (dev.deviceOverlap ? "ok" : "not"));
		printf(" num. of multiprocessors : %d\n", dev.multiProcessorCount);
		printf(" kernel execution timeout : %s\n", (dev.kernelExecTimeoutEnabled ? "on" : "off"));
		printf(" integrated : %s\n", (dev.integrated ? "on" : "off"));
		printf(" host memory mapping : %s\n", (dev.canMapHostMemory ? "on" : "off"));

		printf(" compute mode : ");
		if(dev.computeMode == cudaComputeModeDefault) printf("default mode (multiple threads can use) \n");
		else if(dev.computeMode == cudaComputeModeExclusive) printf("exclusive mode (only one thread will be able to use)\n");
		else if(dev.computeMode == cudaComputeModeProhibited) printf("prohibited mode (no threads can use)\n");
		
	}

	printf("Device with Maximum GFLOPS : %d\n", cutGetMaxGflopsDeviceId());
}


}   // extern "C"
