/*! @file rx_ps.cuh

	@brief CUDAɂSPH
 
	@author Makoto Fujisawa
	@date 2009-08
*/

#ifndef _RX_SSM_CUH_
#define _RX_SSM_CUH_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_ssm_kernel.cuh"


//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
//-----------------------------------------------------------------------------
// SSM
//-----------------------------------------------------------------------------
void CuInitDepthMap(float* dDMap, int nx, int ny);
void CuCreateDepthMap(float* dDMap, float* dPrtPos, float* dPrtRad, int pnum, float* tr, float* pmv, 
					  int W, int H, float dw, float dh, int nx, int ny);
void CuDepthSmoothing(float* dDMap, int nx, int ny, int n_filter, float *binomials, float zmax);
void CuDetectSilhouetteEdgeVertex(float* dDMap, int nx, int ny, float dw, float dh, float zmax, 
								  float* dPrtPos, float* dPrtRad, int pnum, int W, int H, 
								  rxVPacke &dEdge, rxVPackf &dEdgeVrts, rxVPackf &dBackEdgeVrts, rxSSGridG* dMGrid, 
								  int num_nv, int &num_edge, int &num_front_edge, int &num_back_edge);
void CuCalNodeVertex(float* dDMap, int nx, int ny, float dw, float dh, float zmax, 
					 rxVPackf &dNodeVrts, rxSSGridG* dMGrid, int &num_node_vertex);
void CuCalMesh(rxSSGridG* dMGrid, int nx, int ny, float dw, float dh, float zmax, uint* dTriNum, uint* dTriNumScan, 
			   rxVPackf &dBack2Vrts, float* dVrts, int num_vrts, int &num_back2_vrts, uint* &dTriArray, int &num_mesh);
void CuSilhouetteSmoothing(float* dVrts, int num_vrts, int num_node_vrts, uint* dTriArray, int num_mesh, int n_iter);
void CuTransformBack(float* dSSVrts, int num_vrts, float* mvq, float* q, int W, int H, float* dVrts);
void CuCalVertexNormal(float* dVrts, int num_vrts, uint* dTriArray, int num_mesh, float* dNrms);

void CuSetParameters(rxSsmParams *hostParams);
void CuCreateScanPlan(int num);

void CuInitVPackf(rxVPackf &pack, int size);
void CuCleanVPackf(rxVPackf &pack);
void CuInitVPacke(rxVPacke &pack, int size);
void CuCleanVPacke(rxVPacke &pack);

void CuInitTable(void);
void CuCleanTable(void);


//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
void CuInit(int argc, char **argv);
void CuSetDevice(int id);

void CuDeviceProp(void);

void CuAllocateArray(void **devPtr, int size);
void CuFreeArray(void *devPtr);

void CuCopyArrayD2D(void *dDst, void *dSrc, int size);
void CuThreadSync();

void CuCopyArrayFromDevice(void* host, const void* device, unsigned int vbo, int size);
void CuCopyArrayToDevice(void* device, const void* host, int offset, int size);
void CuRegisterGLBufferObject(unsigned int vbo);
void CuUnregisterGLBufferObject(unsigned int vbo);
void *CuMapGLBufferObject(uint vbo);
void CuUnmapGLBufferObject(uint vbo);

}


#endif // #ifdef _RX_SSM_CUH_