/*!
  @file rx_ssmesh_common.h
	
  @brief Screen Space Mesh쐬NXʃwb_
 		- M. Muller et al., Screen space meshes, SCA2007, 2007. 

  @author Makoto Fujisawa
  @date 2011-05
*/
// FILE --rx_ssmesh_common.h--


#ifndef _RX_SSMESH_COMMON_H_
#define _RX_SSMESH_COMMON_H_

#pragma warning (disable: 4244)


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cmath>
#include <cstdlib>

#include <iostream>
#include <fstream>

// STL
#include <vector>
#include <string>

// OpenGL
#include <GL/glut.h>

// f[^\
#include "rx_utility.h"
#include "rx_matrix.h"

// Marching Squaree[u
#include "rx_ms_tables.h"

#include "rx_ssm_kernel.cuh"

using namespace std;



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
#define RXREAL float
#define RXREAL3 float3
#define RXREAL2 float2



//! Screen Spaceł̃p[eBN
struct rxSSParticle
{
	Vec3 xp;	//!< XN[Xy[Xł̒SW
	Vec3 rp;	//!< XN[Xy[Xł̔a
};

//! Screen Space ł̒_
struct rxSSVertex
{
	Vec3 pos;		//!< Screen SpaceWl
	RXREAL depth;	//!< fvXl
	int edge;		//!< GbW_ł邩ǂ

	Vec3 avg_pos;	//!< אڒ_ύWl(֊sp)
	RXREAL avg_num;	//!< אڒ_(֊sp)

	rxSSVertex(Vec3 p, int e = 0) : pos(Vec3(p[0], p[1], p[2])), depth(p[2]), edge(e) {}
};

//! ֊sGbW
struct rxSSEdge
{
	Vec3 x0, x1;	//!< [_W
	RXREAL depth;	//!< GbWfvXl
	RXREAL d0, d1;	//!< [_fvXl
	int xy;			//!< GbW(x:0, y:1)
	int silhouette;	
	int front_vertex;	//!< GbW_̃CfbNX
	RXREAL dx;			//!< fvXl[_GbW_܂ł̋
};

//! bVpObh
struct rxSSGrid
{
	int i, j;
	int node_vrts[4];	//!< m[h_CfbNX
	int num_nv;			//!< m[h_
	int edge_vrts[4];	//!< GbW_(front vertex)
	int num_ev;			//!< GbW_(front vertex)
	int back_vrts[6];	//!< GbW_(back vertex, back-2 vertex)
	int num_bv;			//!< GbW_(back vertex)

	RXREAL node_depth[4];	//!< m[h̃fvXl

	int table_index0;	//!< fobOp:bV̂߂̃CfbNXl
	int table_index1;	//!< fobOp:bV̂߂̃CfbNXl
	int mesh_num;		//!< fobOp:bV
	int mesh[6];		//!< fobOp:bVCfbNX
	int back2;			//!< fobOp
	int v[14];			//!< fobOp
};


// m[h_
// 3 - 2
// |   |
// 0 - 1
	 
// GbW_
// - 2 -
// 3   1
// - 0 -




//-----------------------------------------------------------------------------
// ֐
//-----------------------------------------------------------------------------
/*!
 * Nv
 * @param[inout] x l
 * @param[in] a,b [a,b]ŃNv
 */
template<class T> 
inline void RX_CLAMP(T &x, const T &a, const T &b){ x = (x < a ? a : (x > b ? b : x)); }

inline void Swap(Vec4 &a, Vec4 &b)
{
	Vec4 tmp = a;
	a = b;
	b = tmp;
}


inline RXREAL GetDepthInterpT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap)
{
	int ix0 = (int)(x/dx);
	int ix1 = ix0+1;
	double ddx = x/dx-ix0;

	RX_CLAMP(ix0, 0, nx-1);
	RX_CLAMP(ix1, 0, nx-1);

	int iy0 = (int)(y/dx);
	int iy1 = iy0+1;
	double ddy = y/dy-iy0;

	RX_CLAMP(iy0, 0, ny-1);
	RX_CLAMP(iy1, 0, ny-1);

	RXREAL d00 = dmap[iy0*nx+ix0];
	RXREAL d01 = dmap[iy0*nx+ix1];
	RXREAL d10 = dmap[iy1*nx+ix0];
	RXREAL d11 = dmap[iy1*nx+ix1];

	return (d00*(1.0-ddx)+d01*ddx)*(1.0-ddy)+(d10*(1.0-ddx)+d11*ddx)*ddy;
}


inline RXREAL GetDepthNearestT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap)
{
	int ix = (int)(x/dx);
	RX_CLAMP(ix, 0, nx-1);

	int iy = (int)(y/dx);
	RX_CLAMP(iy, 0, ny-1);

	return dmap[iy*nx+ix];
}

// 񕪒T(1D)
inline double rtbis(double func(const double), const double x1, const double x2, const double xacc)
{
	const int JMAX = 40;
	double dx, f, fmid, xmid, rtb;

	f = func(x1);
	fmid = func(x2);
	if(f*fmid >= 0.0) return 0.0;

	rtb = f < 0.0 ? (dx = x2-x1, x1) : (dx = x1-x2, x2);
	for(int j = 0; j < JMAX; ++j){
		dx *= 0.5;
		xmid = rtb+dx;

		fmid = func(xmid);

		if(fmid <= 0.0){
			rtb = xmid;
		}

		if(fabs(dx) < xacc || fmid == 0.0){
			return rtb;
		}
	}
	
	return 0.0;
}


/*!
 * OpenGLϊsrxMatrix4ɕϊ
 *  column major  row major ɕϊrxMatrix4Ɋi[
 * @param[in] m OpenGLϊs(	glGetDoublev(GL_PROJECTION_MATRIX, m); ȂǂŎ擾)
 * @return ϊ̍s
 */
inline rxMatrix4 GetMatrixGL(double *m)
{
	return rxMatrix4(m[0], m[4], m[8],  m[12], 
					 m[1], m[5], m[9],  m[13], 
					 m[2], m[6], m[10], m[14], 
					 m[3], m[7], m[11], m[15]);
}


/*!
 * 2iɕϊ
 * @param[in] x ̐
 * @param[in] bit 2i
 * @return 2i
 */
inline string GetBitArray(int x, int bit)
{
    string s;
    s.resize(bit, '0');
    for(int i = 0; i < bit; ++i){
        s[bit-i-1] = ((x >> i) & 0x01) ? '1' : '0';
    }
    return s;
}

/*!
 * fvXl[0,1]ɕϊ
 * @param[in] depth fvXl
 * @param[in] w divisionl
 * @return [0,1]̒l
 */
inline double RX_DEPTH2COLORf(RXREAL depth)
{
	if(depth == RX_FEQ_INF){
		depth = 1.0;
	}
	else{
		depth *= 0.5;
	}
	return 1.0-RX_CLAMP(depth, 0.0, 1.0);
}
inline unsigned char RX_DEPTH2COLOR(RXREAL d){ return (unsigned char)(RX_DEPTH2COLORf(d)*255); }


/*!
 * pbinominalWvZ
 * @param[in] b 
 * @return binominalW
 */
static vector< vector<double> > CalBinomials(int b)
{
	vector< vector<double> > bs;
	vector<double> f, tmp;
	f.resize(b+1);
	tmp.resize(b+1);

	bs.resize(b+1);

	double a = 1.0;

	for(int i = 0; i < b+1; ++i){
		f[i]   = (i == 0 ? 1 : 0);
		tmp[i] = (i == 0 ? 1 : 0);	
	}

	for(int k = 0; k < b+1; ++k){
		for(int i = 1; i < k+1; ++i){
			tmp[i] = f[i-1]+f[i];
		}

		for(int i = 1; i < k+1; ++i){
			f[i] = tmp[i];
		}

		bs[k].resize(k+1);
		for(int i = 0; i < k+1; ++i){
			bs[k][i] = f[i]*a;
		}

		a *= 0.5;
	}

	return bs;
}

/*!
 * pbinominalWvZ(1z, tB^ɗp镨̂)
 * @param[in] r tB^(b = 2*r+1)
 * @return binominalW
 */
static vector<float> CalBinomialsForFilter(int r)
{
	int b = 2*r+1;
	vector<float> bs;
	vector<float> f, tmp;
	f.resize(b+1);
	tmp.resize(b+1);

	bs.resize((r+1)*(r+1));

	float a = 1.0;

	for(int i = 0; i < b+1; ++i){
		f[i]   = (i == 0 ? 1.0f : 0.0f);
		tmp[i] = (i == 0 ? 1.0f : 0.0f);	
	}

	int c = 0;
	for(int k = 0; k < b+1; ++k){
		for(int i = 1; i < k+1; ++i){
			tmp[i] = f[i-1]+f[i];
		}

		for(int i = 1; i < k+1; ++i){
			f[i] = tmp[i];
		}

		if(!(k%2)){
			for(int i = 0; i < k+1; ++i){
				bs[c++] = f[i]*a;
			}
		}

		a *= 0.5;
	}

	return bs;
}

/*!
 * Ɖ~̌(2D)
 * @param[in] A,B ̗[_W
 * @param[in] C ~̒S
 * @param[in] r ~̔a
 * @param[out] P _W
 * @return _
 */
static int LineCircleIntersection(const Vec2 &A, const Vec2 &B, const Vec2 &C, const double &r, Vec2 P[2], double t[2])
{
	double rr = r*r;
	Vec2 AC = C-A;
	Vec2 BC = C-B;

	Vec2 v = B-A;
	double l = norm(v);
	v /= l;

	double td = dot(v, AC);
	Vec2 D = A+td*v;
	double dd = norm2(D-C);

	if(dd < rr){
		double dt = sqrt(rr-dd);

		double da = rr-norm2(AC);
		double db = rr-norm2(BC);

		int inter = 0;
		double t1 = td-dt;
		double t2 = td+dt;
		if(t1 >= 0 && t1 <= l){
			P[inter] = A+t1*v;
			t[inter] = t1;
			inter++;
		}
		if(t2 >= 0 && t2 <= l){
			P[inter] = A+t2*v;
			t[inter] = t2;
			inter++;
		}

		return inter;
	}
	else{
		return 0;
	}
}

#endif // #ifndef _RX_SSMESH_COMMON_H_
