/*! @file rx_time.h
	
	@brief Ԍv
 
	@author Makoto Fujisawa
	@date 2011
*/
// FILE --rx_time.h--

#ifndef _RX_TIME_H_
#define _RX_TIME_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <ctime>
#include <cmath>
#include <cstdio>

#include <iostream>
#include <sstream>
#include <iomanip>

// STL
#include <vector>
#include <string>
#include <map>

using namespace std;


//-----------------------------------------------------------------------------
// MARK:Ԍv
//-----------------------------------------------------------------------------
#define RX_USE_MM
#ifdef WIN32
	#include <windows.h>
 
    #ifdef RX_USE_MM
    #include <mmsystem.h>
    #pragma comment (lib, "winmm.lib")
    #endif
 #endif
 
#ifdef WIN32
    #ifdef RX_USE_MM
        #define RXTIME DWORD
        #define RXGETTIME timeGetTime
        #define RXTIME2SEC 1.0e-3
        //#define RXTIME2SEC 1.0
    #else
        #define RXTIME DWORD
        #define RXGETTIME GetTickCount
        #define RXTIME2SEC 1.0e-3
        //#define RXTIME2SEC 1.0
    #endif
#else
    #define RXTIME clock_t
    #define RXGETTIME clock
    #define RXTIME2SEC (1.0/CLOCKS_PER_SEC)
#endif


//-----------------------------------------------------------------------------
// ^C}[NX
//-----------------------------------------------------------------------------
class rxTimer
{
    RXTIME m_tStart, m_tEnd;
    vector<double> m_vTimes;
    vector<string> m_vComments;
 
public:
    //! RXgN^
    rxTimer(){}
 
    //! fXgN^
    ~rxTimer(){}
 
    //! vJn
    void Start(void)
    {
        m_tStart = RXGETTIME();
    }
 
    //! v
    void Split(const string &cmnt = "", bool restart = false)
    {
        m_tEnd = RXGETTIME();
 
        double time = (double)(m_tEnd-m_tStart)*RXTIME2SEC;
        m_vTimes.push_back(time);
        m_vComments.push_back(cmnt);
 
        if(restart) m_tStart = RXGETTIME();
    }
 
    //! vI
    void Stop(const string &cmnt = "")
    {
        m_tEnd = RXGETTIME();
 
        double time = (double)(m_tEnd-m_tStart)*RXTIME2SEC;
        m_vTimes.push_back(time);
        m_vComments.push_back(cmnt);
 
        m_tStart = m_tEnd = 0;
    }
 
    //! Zbg
    void Reset(void)
    {
        m_vTimes.clear();
        m_vComments.clear();
        m_tStart = m_tEnd = 0;
    }
 
    // ŌɋL^ꂽԂ폜
    void PopBackTime(void)
    {
        m_vTimes.pop_back();
        m_vComments.pop_back();
    }
 
    //! ԂZbg(̌v@ŌvʂȂ)
    void SetTime(const double &t, const string &cmnt = "")
    {
        m_vTimes.push_back(t);
        m_vComments.push_back(cmnt);
    }
 
    //! Ԃ̎擾
    double GetTime(int i)
    {
        if(i >= (int)m_vTimes.size()) return 0.0;
 
        return m_vTimes[i];
    }
 
    //! L^ꂽԐ̎擾
    int GetTimeNum(void)
    {
        return (int)m_vTimes.size();
    }
 
    //! L^ꂽԂʏo
    double Print(void)
    {
        int m = 0, mi;
        if(m_vTimes.empty()){
            return 0.0;
        }
        else{
            // vԂvZ
            double total = 0.0;
            for(int i = 0; i < (int)m_vTimes.size(); ++i){
                mi = (int)m_vComments[i].size();
                if(mi > m) m = mi;
 
                total += m_vTimes[i];
            }
 
            SetTime(total, "total");
        }
 
        int cur_p = cout.precision();
        cout.precision(3);
        cout.setf(ios::fixed);
        for(int i = 0; i < (int)m_vTimes.size(); ++i){
            string spc = RXFunc::GenSpace(m-(int)m_vComments[i].size());
            cout << m_vComments[i] << spc << " : " << m_vTimes[i] << endl;
        }
        cout.unsetf(ios::fixed);
        cout.precision(cur_p);
 
        double t = m_vTimes.back();
 
        PopBackTime(); // i[vԂ̌vZɔč폜
 
        return t;
    }
 
    //! L^ꂽԂ𕶎ɏo
    double PrintToString(string &str)
    {
        int m = 0, mi;
        if(m_vTimes.empty()){
            return 0.0;
        }
        else{
            // vԂvZ
            double total = 0.0;
            for(int i = 0; i < (int)m_vTimes.size(); ++i){
                mi = (int)m_vComments[i].size();
                if(mi > m) m = mi;
 
                total += m_vTimes[i];
            }
 
            SetTime(total, "total");
        }
 
        stringstream ss;
        ss.precision(3);
        ss.setf(ios::fixed);
        
        int n = (int)m_vTimes.size();
        for(int i = 0; i < n; ++i){
            string spc = RXFunc::GenSpace(m-(int)m_vComments[i].size());
            ss << m_vComments[i] << spc << " : " << m_vTimes[i] << "\n";
        }
 
        ss << "\n";
        str = ss.str();
 
        double t = m_vTimes.back();
 
        PopBackTime(); // i[vԂ̌vZɔč폜
 
        return t;
    }
};



//-----------------------------------------------------------------------------
// ^C}[NX(ϒl)
//  - rxTimerŌvԂXvbgƂɒ~ςĕϒl߂
//-----------------------------------------------------------------------------
class rxTimerAvg
{
public:
	// Ԃƌv
	struct rxTimeAndCount
	{
		double time;
		int count;
		int idx;
	};

	// Ԃƌv񐔂𕶎Ɗ֘AÂ}bv
	typedef map<string, rxTimeAndCount> RXMAPTC;

private:
	rxTimer m_Tmr;		//!< ԌvNX
	RXMAPTC m_TimeMap;	//!< Ԃƌv񐔂𕶎Ɗ֘AÂ}bv

public:
    //! RXgN^
	rxTimerAvg()
	{
		Clear();
		ResetTime();
		ClearTime();
	}

	/*!
	 * ׂăNA
	 */
	void Clear(void)
	{
		m_TimeMap.clear();
	}

	/*!
	 * ~ώԂ̏
	 */
	void ClearTime(void)
	{
		for(RXMAPTC::iterator it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
			it->second.time = 0.0;
			it->second.count = 1;
			//it->second.idx = -1;
		}
	}

	/*!
	 * Zbg
	 */
	void ResetTime(void)
	{
		m_Tmr.Reset();
		m_Tmr.Start();
	}

	/*!
	 * v
	 * @param[in] cmnt Ԓ~ϗp̖O
	 */
	void Split(const string &cmnt)
	{
		RXMAPTC::iterator i = m_TimeMap.find(cmnt);
		
		m_Tmr.Stop();
		if(i == m_TimeMap.end()){
			
			m_TimeMap[cmnt].time = m_Tmr.GetTime(0);
			m_TimeMap[cmnt].count = 1;
			m_TimeMap[cmnt].idx = m_TimeMap.size()-1;
		}
		else{
			m_TimeMap[cmnt].time += m_Tmr.GetTime(0);
			m_TimeMap[cmnt].count++;
		}
		m_Tmr.Reset();
		m_Tmr.Start();
	}

	/*!
	 * Ԃ̎擾
	 * @return 
	 */
	double GetTotalTime(void)
	{
		if(m_TimeMap.empty()){
			m_Tmr.Stop();
			return m_Tmr.GetTime(0);
		}
		else{
			double total = 0.0;
			for(RXMAPTC::iterator it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
				total += it->second.time/it->second.count;
			}

			return total;
		}
	}

	/*!
	 * L^ꂽԂʏo
	 */
	void Print(void)
	{
		int m = 0, mi;
		double total = 0.0;
		for(RXMAPTC::iterator it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
			mi = (int)it->first.size();
			if(mi > m) m = mi;

			total += it->second.time/it->second.count;
		}

        int cur_p = cout.precision();
        cout.precision(3);
        cout.setf(ios::fixed);
		for(RXMAPTC::iterator it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
			string spc = RXFunc::GenSpace( m-(int)(it->first.size()) );

			cout << it->first << spc << " : " << (it->second.time/it->second.count) << "[s]" << endl;
		}
        cout.unsetf(ios::fixed);
        cout.precision(cur_p);

		string spc = RXFunc::GenSpace(m-5);
		cout << "total" << spc << " : " << total << endl;
	}

	/*!
	 * L^ꂽԂ𕶎ɏo
	 * @param[out] str o͕
	 */
	void PrintToString(string &str)
	{
		int m = 0, mi;
		double total = 0.0;
		for(RXMAPTC::iterator it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
			mi = (int)it->first.size();
			if(mi > m) m = mi;

			total += it->second.time/it->second.count;
		}

		stringstream ss;
		ss.precision(3);
		ss.setf(ios::fixed);
		for(int i = 0; i < (int)m_TimeMap.size(); ++i){
			RXMAPTC::iterator it = m_TimeMap.begin();

			for(it = m_TimeMap.begin(); it != m_TimeMap.end(); ++it){
				if(it->second.idx == i){
					break;
				}
			}

			string spc = RXFunc::GenSpace(m-(int)(it->first.size()));
			ss << it->first << spc << " : " << (it->second.time/it->second.count) << "[s]\n";
		}

		string spc = RXFunc::GenSpace(m-5);
		ss << "total" << spc << " : " << total << "[s]\n";

		str = ss.str();
	}
};


#endif // #ifndef _RX_COMMON_H_